# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module docstring."""
from datetime import datetime

import ipywidgets as widgets

from .._version import VERSION
from ..common.utility.package import delayed_import

# from ..context.azure.sentinel_core import MicrosoftSentinel
from .ce_common import (
    ITEM_LIST_LAYOUT,
    get_or_create_mpc_section,
    get_wgt_ctrl,
    print_debug,
)
from .comp_edit import CEItemsBase, CompEditDisplayMixin
from .mp_config_control import MpConfigControls

__version__ = VERSION
__author__ = "Ian Hellen"

ms_sentinel = delayed_import("msticpy.context.azure.sentinel_core", "MicrosoftSentinel")


# pylint: disable=too-many-ancestors
class CEAzureSentinel(CEItemsBase):
    """Microsoft Sentinel Workspaces editor component."""

    _DESCRIPTION = "Microsoft Sentinel workspace settings"
    _COMP_PATH = "AzureSentinel.Workspaces"
    _HELP_TEXT = """
    Supply the parameters for your Microsoft Sentinel workspaces here.<br>

    You can use the URL import to quickly fetch Microsoft Sentinel Workspace
    settings. In your browser, navigate to your Sentinel workspace in the
    Azure portal (e.g. the Overview page). Copy the URL and paste into the
    <b>Portal URL</b> text box, then click on <b>Import from URL</b>.

    You can also add partial information (such as a Workspace ID) and use
    the <b>Resolve Settings</b> button to fetch the other settings from
    the Azure Resource Graph.

    Note you must be authenticated to Azure for either of these to work.

    You can also get all of these settings (apart from 'TenantID') from your
    workspace portal manually.
    Navigate to "Settings" (on the left side of the screen), then click the
    "Workspace Settings" tab near the top of the page.

    The Tenant ID is the only value missing from this settings page. If you cannot
    find this you can use the following code to try to look it up:
    <pre>
    from msticpy.common.config.ce_common import get_def_tenant_id
    get_def_tenant_id("{workspace-subscription-id}")
    </pre>

    The name that you use for workspace (the "Name" text box) does not have to
    be the same as the Workspace name. You can use any helpful name that you
    like. You can even have the same workspace included multiple times with
    different names.

    There is a special case of the workspace name "Default". Normally, when you
    connect to a workspace you need to supply the name corresponding to the
    workspace entry in your settings. If you give a workspace the name "Default"
    it will be picked if you do not specify a name. Note, the "Default" entry
    can be a duplicate of another named entry.
    """
    _HELP_URI = {
        "Connecting to a Microsoft Sentinel Workspace": (
            "https://msticpy.readthedocs.io/en/latest/data_acquisition/"
            + "DataProviders.html#connecting-to-an-azure-sentinel-workspace"
        ),
        "Help on this tab": (
            "https://msticpy.readthedocs.io/en/latest/getting_started/"
            + "SettingsEditor.html#azure-sentinel-workspaces"
        ),
    }

    def __init__(self, mp_controls: MpConfigControls):
        """
        Initialize an instance of CEAzureSentinel.

        Parameters
        ----------
        mp_controls : MpConfigControls
            The config/controls/settings database

        """
        super().__init__(mp_controls)

        get_or_create_mpc_section(self.mp_controls, self._COMP_PATH)
        self.select_item.options = self._get_select_opts()
        self.select_item.layout = ITEM_LIST_LAYOUT["layout"]
        self.select_item.style = ITEM_LIST_LAYOUT["style"]
        self.select_item.description = "Workspaces"
        self.select_item.observe(self._select_provider, names="label")

        prov_name = self.select_item.label
        self.edit_ctrls = _get_ws_ctrls(prov_name, self.mp_controls, self._COMP_PATH)
        self.edit_ctrls.children[0].value = prov_name or ""
        self.btn_set_default = widgets.Button(
            description="Set as default",
            tooltip="Set the current workspace as your default.",
        )
        btn_resolve = widgets.Button(
            description="Resolve settings",
            tooltip="Try to resolve any missing settings using the Azure resource graph.",
        )
        self.btn_imp_url = widgets.Button(
            description="Import from URL",
            tooltip="Try to lookup workspace settings from the Sentinel portal URL.",
        )
        self.txt_imp_url = widgets.Text(
            description="Portal URL",
            # Tooltip not yet supported in widgets.Text
            # tooltip="Paste in the URL from the Sentinel Azure portal to fetch settings.",
        )
        self.ws_btns = widgets.HBox([self.btn_set_default, btn_resolve])
        self.url_imp_ctrls = widgets.HBox([self.txt_imp_url, self.btn_imp_url])
        if prov_name:
            self.edit_frame.children = [
                self.edit_ctrls,
                self.ws_btns,
                self.url_imp_ctrls,
            ]
        else:
            self.edit_frame.children = [self.edit_ctrls]
        self.edit_buttons.btn_del.on_click(self._del_item)
        self.edit_buttons.btn_add.on_click(self._add_item)
        self.edit_buttons.btn_save.on_click(self._save_item)
        self.btn_set_default.on_click(self._set_default)
        btn_resolve.on_click(self._resolve_settings)
        self.btn_imp_url.on_click(self._imp_ws_from_url)
        self.current_workspace = prov_name

    @property
    def _current_path(self):
        return f"{self._COMP_PATH}.{self.select_item.label}"

    def _get_select_opts(self):
        """Get provider options to populate select list."""
        provs = self.mp_controls.get_value(self._COMP_PATH)
        return [(val, idx) for idx, val in enumerate(provs.keys())]

    def _select_provider(self, change):
        """Select a provider from the list."""
        prov_name = change.get("new")
        self.edit_ctrls = _get_ws_ctrls(prov_name, self.mp_controls, self._COMP_PATH)
        self.edit_ctrls.children[0].value = prov_name
        self.current_workspace = prov_name
        self.edit_frame.children = [self.edit_ctrls, self.ws_btns, self.url_imp_ctrls]
        self.mp_controls.populate_ctrl_values(self.current_workspace)

    def _add_item(self, btn):
        """Add an item."""
        del btn
        # Generate a "unique" temp name
        new_provider = f"NewWorkspace_{datetime.now().strftime('%f')[:4]}"
        while new_provider in dict(self.select_item.options):
            new_provider = f"NewWorkspace_{datetime.now().strftime('%f')[:4]}"
        _get_ws_ctrls(new_provider, self.mp_controls, self._COMP_PATH)
        self.mp_controls.save_ctrl_values(f"{self._COMP_PATH}.{new_provider}")
        current_options = self._get_select_opts()
        self.select_item.options = current_options
        self.select_item.label = new_provider

    def _del_item(self, btn):
        """Delete an item."""
        del btn
        self.mp_controls.del_value(self._current_path)
        remaining_opts = self._get_select_opts()
        self.select_item.options = remaining_opts
        if remaining_opts:
            self.select_item.label = remaining_opts[-1][0]

    def _save_item(self, btn):
        """Save the current item."""
        del btn
        # Use may have edited the workspace name.
        # If so save the existing one and rename the paths to match new name
        edited_provider_name = self.edit_ctrls.children[0].value
        self.mp_controls.save_ctrl_values(self._current_path)
        if edited_provider_name != self.current_workspace and self.current_workspace:
            self.mp_controls.rename_path(
                f"{self._COMP_PATH}.{self.current_workspace}",
                f"{self._COMP_PATH}.{edited_provider_name}",
            )
            self.select_item.options = self._get_select_opts()
            self.select_item.label = edited_provider_name
        valid, status = _validate_ws(
            edited_provider_name, self.mp_controls, self._COMP_PATH
        )
        if not valid:
            self.set_status(status)

    def _set_default(self, btn):
        """Set selected item to be the Default entry."""
        del btn
        src_wkspace = self.current_workspace
        if src_wkspace == "Default" or not src_wkspace:
            return
        src_path = f"{self._COMP_PATH}.{src_wkspace}"
        src_settings = self.mp_controls.get_value(src_path)
        def_path = f"{self._COMP_PATH}.Default"
        # save the src values to the Default key and refresh the select_item list
        self.mp_controls.set_value(def_path, src_settings)
        print_debug("Current options", self._get_select_opts())
        self.select_item.options = self._get_select_opts()
        self.select_item.label = "Default"

    def _imp_ws_from_url(self, btn):
        """Import workspace settings from a portal URL."""
        del btn
        url = self.txt_imp_url.value
        if not url:
            self.set_status("Please paste portal URL into ")
            return
        self._update_settings(ms_sentinel().get_workspace_details_from_url(url))

    def _update_settings(self, ws_settings):
        """Update current controls with workspace settings."""
        if not ws_settings:
            self.set_status("Could not resolve workspace from URL")
            return
        ws_name = next(iter(ws_settings))
        _get_named_control(self.edit_ctrls, "Name").value = ws_name
        for setting, value in ws_settings[ws_name].items():
            ctrl = _get_named_control(self.edit_ctrls, setting)
            if ctrl is None or ctrl.value:
                # don't overwrite existing settings
                continue
            ctrl.value = value

    def _resolve_settings(self, btn):
        """Resolve missing settings for workspace."""
        del btn
        subscription_id = _get_named_control(self.edit_ctrls, "SubscriptionId").value
        workspace_id = _get_named_control(self.edit_ctrls, "WorkspaceId").value
        workspace_name = _get_named_control(self.edit_ctrls, "WorkspaceName").value
        resource_group = _get_named_control(self.edit_ctrls, "ResourceGroup").value
        if not (workspace_id or workspace_name):
            self.set_status(
                "Need at least WorkspaceId or WorkspaceName to lookup settings."
            )
            return
        if workspace_id:
            self._update_settings(
                ms_sentinel().get_workspace_settings(workspace_id=workspace_id)
            )
        else:
            self._update_settings(
                ms_sentinel().get_workspace_settings_by_name(
                    workspace_name=workspace_name,
                    subscription_id=subscription_id,
                    resource_group=resource_group,
                )
            )

    def _select_labels(self):
        return [label for label, _ in self.select_item.options]


_TEXT_PARAMS = {
    "layout": widgets.Layout(width="70%"),
    "style": {"description_width": "150px"},
}


def _get_ws_ctrls(workspace, mp_controls, conf_path):
    """Return or create the control set for a workspace."""
    defn_path = f"{conf_path}.Default"
    prov_defn = mp_controls.get_defn(defn_path)

    ctrls = [widgets.Text(description="Name", **_TEXT_PARAMS)]
    if not workspace:
        return widgets.VBox(ctrls, layout=CompEditDisplayMixin.no_border_layout("95%"))
    prov_path = f"{conf_path}.{workspace}"
    for setting in prov_defn:
        wgt = get_wgt_ctrl(prov_path, setting, mp_controls)
        wgt.layout = _TEXT_PARAMS["layout"]
        wgt.style = _TEXT_PARAMS["style"]
        ctrls.append(wgt)

    return widgets.VBox(ctrls)


def _get_named_control(edit_ctrls, name):
    """Get the control with matching name."""
    try:
        return next(ctrl for ctrl in edit_ctrls.children if ctrl.description == name)
    except StopIteration:
        return None


def _validate_ws(workspace, mp_controls, conf_path):
    """Validate the settings for a workspace."""
    defn_path = f"{conf_path}.Default"

    if not workspace:
        return False, "Workspace name must be supplied."
    prov_path = f"{conf_path}.{workspace}"
    results = mp_controls.validate_setting(path=prov_path, defn_path=defn_path)
    status = []
    validated = True
    for valid, mssg in results:
        if not valid:
            validated = False
            status.append(mssg)
    return validated, "  ".join(status)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Timeline values Bokeh plot."""
from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import attr
import pandas as pd
from bokeh.io import output_notebook, show
from bokeh.layouts import column
from bokeh.models import (  # type: ignore[attr-defined]
    ColumnDataSource,
    HoverTool,
    LayoutDOM,
    Legend,
)
from bokeh.plotting import figure, reset_output

from .._version import VERSION
from ..common.data_utils import ensure_df_datetimes
from ..common.utility import check_kwargs, export
from .figure_dimension import bokeh_figure
from .timeline_common import (
    TIMELINE_HELP,
    calc_auto_plot_height,
    check_df_columns,
    create_data_grouping,
    create_range_tool,
    create_tool_tips,
    get_ref_event_time,
    get_tick_formatter,
    plot_ref_events,
    plot_ref_line,
)

__version__ = VERSION
__author__ = "Ian Hellen"

# wrap figure function to handle v2/v3 parameter renaming
figure = bokeh_figure(figure)  # type: ignore[assignment, misc]


@attr.s(auto_attribs=True)
class PlotParams:
    """Plot params for time_duration."""

    time_column: Optional[str] = None
    height: Optional[int] = None
    width: int = 900
    title: Optional[str] = None
    yaxis: bool = True
    range_tool: bool = True
    group_by: Optional[str] = None
    legend: Optional[str] = None
    xgrid: bool = True
    ygrid: bool = False
    hide: bool = False
    color: str = "navy"
    kind: Union[str, List[str]] = "vbar"
    ylabel_cols: Iterable[str] = attr.Factory(list)
    ref_event: Optional[Any] = None
    ref_time: Optional[datetime] = None
    ref_events: Optional[pd.DataFrame] = None
    ref_col: Optional[str] = None
    ref_time_col: Optional[str] = None
    ref_times: Optional[List[Tuple[datetime, str]]] = None
    source_columns: List = []

    @classmethod
    def field_list(cls) -> List[str]:
        """Return field names as a list."""
        return list(attr.fields_dict(cls).keys())


# pylint: disable=invalid-name, too-many-locals, too-many-statements, too-many-branches
@export  # noqa: C901, MC0001
def display_timeline_values(  # noqa: C901, MC0001
    data: pd.DataFrame,
    value_column: str = None,
    time_column: str = "TimeGenerated",
    source_columns: list = None,
    **kwargs,
) -> LayoutDOM:
    """
    Display a timeline of events.

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame as a single data set or grouped into individual
        plot series using the `group_by` parameter
    time_column : str, optional
        Name of the timestamp column
        (the default is 'TimeGenerated')
    value_column : str
        The column name holding the value to plot vertically
    source_columns : list, optional
        List of default source columns to use in tooltips
        (the default is None)

    Other Parameters
    ----------------
    x : str, optional
        alias of `time_column`
    y : str, optional
        alias of `value_column`
    value_col : str, optional
        alias of `value_column`
    title : str, optional
        Title to display (the default is None)
    ref_event : Any, optional
        Add a reference line/label using the alert time (the default is None)
    ref_time : datetime, optional
        Add a reference line/label using `ref_time` (the default is None)
    ref_label : str, optional
        A label for the `ref_event` or `ref_time` reference item
    group_by : str
        (where `data` is a DataFrame)
        The column to group timelines on
    legend: str, optional
        "left", "right", "inline" or "none"
        (the default is to show a legend when plotting multiple series
        and not to show one when plotting a single series)
    yaxis : bool, optional
        Whether to show the yaxis and labels
    range_tool : bool, optional
        Show the the range slider tool (default is True)
    height : int, optional
        The height of the plot figure
        (the default is auto-calculated height)
    width : int, optional
        The width of the plot figure (the default is 900)
    color : str
        Default series color (default is "navy"). This is overridden by
        automatic color assignments if plotting a grouped chart
    kind : Union[str, List[str]]
        one or more glyph types to plot., optional
        Supported types are "circle", "line" and "vbar" (default is "vbar")
    hide : bool, optional
        If True, create but do not display the plot.
        By default, False.
    ref_events : pd.DataFrame, optional
        Add references line/label using the event times in the dataframe.
        (the default is None)
    ref_time_col : str, optional
        Add references line/label using the this column in `ref_events`
        for the time value (x-axis).
        (this defaults the value of the `time_column` parameter or 'TimeGenerated'
        `time_column` is None)
    ref_col : str, optional
        The column name to use for the label from `ref_events`
        (the default is None)
    ref_times : List[Tuple[datetime, str]], optional
        Add one or more reference line/label using (the default is None)

    Returns
    -------
    LayoutDOM
        The bokeh plot figure.

    """
    # check_kwargs(kwargs, _DEFAULT_KWARGS + _TL_VALUE_KWARGS + ["y"])
    check_kwargs(kwargs, PlotParams.field_list() + ["y", "value_col"])
    value_col = value_column or kwargs.pop("y", kwargs.pop("value_col", None))
    if not value_col:
        raise ValueError("Must supply 'value_column', 'value_col' or 'y' parameter.")
    param = PlotParams(**kwargs)

    if data is None or not isinstance(data, pd.DataFrame) or data.empty:
        print("No data to plot.")
        return figure()

    reset_output()
    output_notebook()
    check_kwargs(kwargs, PlotParams.field_list())
    param = PlotParams(**kwargs)

    plot_kinds = param.kind if isinstance(param.kind, list) else [param.kind]
    param.ref_time_col = param.ref_time_col or time_column or "TimeGenerated"

    ref_time, ref_label = get_ref_event_time(**kwargs)

    if source_columns is None:
        source_columns = [value_col]
    if value_col not in source_columns:
        source_columns.append(value_col)
    check_df_columns(
        data, source_columns + [time_column], TIMELINE_HELP, "display_timeline_values"
    )
    data = ensure_df_datetimes(data=data, columns=time_column)
    graph_df, group_count_df, tool_tip_columns, series_count = create_data_grouping(
        data, source_columns, time_column, param.group_by, param.color
    )

    hover = HoverTool(**(create_tool_tips(data, tool_tip_columns)))

    # Create the Plot figure
    min_time = graph_df[time_column].min()
    max_time = graph_df[time_column].max()
    start_range = min_time - ((max_time - min_time) * 0.1)
    end_range = max_time + ((max_time - min_time) * 0.1)
    height = param.height or calc_auto_plot_height(series_count)

    plot = figure(
        x_range=(start_range, end_range),
        min_border_left=50,
        x_axis_label="Event Time",
        x_axis_type="datetime",
        x_minor_ticks=10,
        y_axis_label=value_col,
        tools=[hover, "xwheel_zoom", "box_zoom", "reset", "save", "xpan"],
        toolbar_location="above",
        title=param.title or "Timeline",
        height=height,
        width=param.width,
    )

    plot.yaxis.visible = param.yaxis
    _set_grid_lines(plot)

    # set the tick datetime formatter
    plot.xaxis[0].formatter = get_tick_formatter()

    # plot groups individually so that we can create an interactive legend
    if param.group_by:
        legend_items = _plot_param_group(
            value_col,
            time_column,
            param,
            plot_kinds,
            graph_df,
            group_count_df,
            plot,
        )

        if param.legend == "inline":
            # Position the inline legend
            plot.legend.location = "top_left"
            plot.legend.click_policy = "hide"
        elif param.legend in {"left", "right"}:
            # Create the legend box outside of the plot area
            ext_legend = Legend(
                items=legend_items,
                location="center",
                click_policy="hide",
                label_text_font_size="8pt",
            )
            plot.add_layout(ext_legend, param.legend)  # type: ignore[arg-type]
    else:
        plot_args = {
            "x": time_column,
            "color": param.color,
            "alpha": 0.7,
            "source": ColumnDataSource(graph_df),
        }
        if "vbar" in plot_kinds:
            plot.vbar(top=value_col, width=4, **plot_args)
        if "circle" in plot_kinds:
            plot.circle(y=value_col, size=4, **plot_args)
        if "line" in plot_kinds:
            plot.line(y=value_col, line_width=2, **plot_args)

    # if we have a reference, plot the time as a line
    if ref_time is not None:
        plot_ref_line(plot, ref_time, ref_label, data[value_col].max())
    elif param.ref_events is not None or param.ref_times is not None:
        plot_ref_events(
            plot=plot,
            ref_events=param.ref_events,
            time_col=param.ref_time_col,
            group_count=series_count,
            ref_col=param.ref_col,
            ref_times=param.ref_times,
        )

    plot_layout: LayoutDOM
    if param.range_tool:
        rng_select = create_range_tool(
            data=graph_df,
            min_time=min_time,
            max_time=max_time,
            plot_range=plot.x_range,  # type: ignore[arg-type]
            width=param.width,
            height=height,
            time_column=time_column,
        )
        plot_layout = column(plot, rng_select)
    else:
        plot_layout = plot

    if not param.hide:
        show(plot_layout)
    return plot_layout


def _set_grid_lines(plot):
    """Set default grid lines."""
    plot.ygrid.minor_grid_line_color = "navy"
    plot.ygrid.minor_grid_line_alpha = 0.1
    plot.ygrid.grid_line_color = "navy"
    plot.ygrid.grid_line_alpha = 0.3
    plot.xgrid.minor_grid_line_color = "navy"
    plot.xgrid.minor_grid_line_alpha = 0.1
    plot.xgrid.grid_line_color = "navy"
    plot.xgrid.grid_line_alpha = 0.3


def _plot_param_group(
    value_col,
    time_column,
    param,
    plot_kinds,
    graph_df,
    group_count_df,
    plot,
) -> List[Tuple[str, Any]]:
    """Plot series groups."""
    legend_items: List[Tuple[str, Any]] = []
    for _, group_id in group_count_df[param.group_by].items():
        first_group_item = graph_df[graph_df[param.group_by] == group_id].iloc[0]
        legend_label = str(first_group_item[param.group_by])
        inline_legend = str(group_id)
        group_color = first_group_item["color"]
        row_source = ColumnDataSource(graph_df[graph_df[param.group_by] == group_id])
        p_series = []
        # create default plot args
        plot_args: Dict[str, Any] = {
            "x": time_column,
            "alpha": 0.7,
            "source": row_source,
        }
        if param.legend != "none":
            plot_args["legend_label"] = inline_legend

        if "vbar" in plot_kinds:
            p_series.append(
                plot.vbar(top=value_col, width=4, color="color", **plot_args)
            )
        if "circle" in plot_kinds:
            p_series.append(
                plot.circle(y=value_col, size=4, color="color", **plot_args)
            )
        if "line" in plot_kinds:
            p_series.append(
                plot.line(
                    y=value_col, line_width=2, line_color=group_color, **plot_args
                )
            )
        if not inline_legend:
            legend_items.append((legend_label, p_series))
    return legend_items



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# Author: Thomas Roccia - @fr0gger_
# --------------------------------------------------------------------------
"""MalwareBazaar TI Provider."""
from __future__ import annotations

from enum import Enum
from typing import Any, ClassVar

import httpx
import pandas as pd
from typing_extensions import Self

from ..._version import VERSION

__version__ = VERSION
__author__ = "Thomas Roccia | @fr0gger_"

_BASE_URL = "https://mb-api.abuse.ch/api/v1/"

_QUERY_OBJECTS_MAPPINGS: dict[str, dict[str, str]] = {
    "hash": {"query": "get_info", "hash": "observable"},
    "tag": {"query": "get_taginfo", "tag": "observable", "limit": "limit"},
    "signature": {"query": "get_siginfo", "signature": "observable", "limit": "limit"},
    "filetype": {"query": "get_file_type", "file_type": "observable", "limit": "limit"},
    "clamav": {"query": "get_clamavinfo", "clamav": "observable", "limit": "limit"},
    "imphash": {"query": "get_imphash", "imphash": "observable", "limit": "limit"},
    "dhash": {"query": "get_dhash_icon", "dhash_icon": "observable", "limit": "limit"},
    "yara": {"query": "get_yarainfo", "yara_rule": "observable", "limit": "limit"},
    "tlsh": {"query": "get_tlsh", "tlsh": "observable", "limit": "limit"},
    "telfhash": {"query": "get_telfhash", "telfhash": "observable", "limit": "limit"},
    "gimphash": {"query": "get_gimphash", "gimphash": "observable", "limit": "limit"},
    "issuerinfo": {"query": "get_issuerinfo", "issuer_cn": "observable"},
    "subjectinfo": {"query": "get_subjectinfo", "subject_cn": "observable"},
    "certificate": {"query": "get_certificate", "serial_number": "observable"},
}


class MBEntityType(Enum):
    """MBEntityType: Enum class for MalwareBazaar entity types."""

    HASH = "hash"
    TAG = "tag"
    SIGNATURE = "signature"
    FILETYPE = "filetype"
    CLAMAV = "clamav"
    IMPHASH = "imphash"
    DHASH = "dhash"
    YARA = "yara"
    TLSH = "tlsh"
    TELFHASH = "telfhash"
    CODESIGNISSUER = "issuerinfo"
    CODESIGNSUBJECT = "subjectinfo"
    CODESIGNSN = "certificate"
    GIMPHASH = "gimphash"


class MBlookup:
    """MBlookup Python Class wrapper for MalwareBazaar API."""

    _SUPPORTED_MB_TYPES: ClassVar[set[MBEntityType]] = {
        MBEntityType.HASH,
        MBEntityType.TAG,
        MBEntityType.SIGNATURE,
        MBEntityType.FILETYPE,
        MBEntityType.CLAMAV,
        MBEntityType.IMPHASH,
        MBEntityType.DHASH,
        MBEntityType.YARA,
        MBEntityType.TLSH,
        MBEntityType.TELFHASH,
        MBEntityType.CODESIGNISSUER,
        MBEntityType.CODESIGNSUBJECT,
        MBEntityType.CODESIGNSN,
        MBEntityType.GIMPHASH,
    }

    def __init__(self: MBlookup, mb_key: str | None = None) -> None:
        """Init function to get the API key if necessary."""
        self.mb_key: str | None = mb_key

    def _make_mb_request(self: Self, data: dict[str, Any]) -> pd.DataFrame:
        """Request to the malware bazaar api."""
        res: httpx.Response = httpx.post(_BASE_URL, data=data, timeout=None)
        res.raise_for_status()
        res_data: dict[str, Any] = res.json()
        if res.is_success:
            return pd.json_normalize(res_data["data"])
        err_msg: str = "Query failed"
        raise httpx.ConnectError(err_msg)

    def lookup_ioc(
        self: Self,
        observable: str,
        mb_type: str,
        limit: int = 10,
    ) -> pd.DataFrame:
        """
        Lookup for IOC in MalwareBazaar.

        Parameters
        ----------
        observable : str
            The observable to lookup. It can be a hash, a signature
        mb_type : str
            The type of the observable. It can be a hash, a signature (refer to MBEntityType).
        limit : int, optional
            The number of results to return, default is 100 or 50 in some cases, by default 10

        Returns
        -------
        pd.DataFrame
            The results of the lookup.

        Raises
        ------
        KeyError
            If invalid IoC type is provided.

        """
        if MBEntityType(mb_type) not in self._SUPPORTED_MB_TYPES:
            err_msg: str = (
                f"Property type {mb_type} not supported."
                " Valid types are "
                ", ".join(x.value for x in MBEntityType.__members__.values())
            )
            raise KeyError(err_msg)

        query_parameters: dict[str, str] = _build_query_string(
            observable,
            mb_type,
            limit,
        )

        return self._make_mb_request(query_parameters)

    def download_sample(self: Self, sha2: str) -> pd.DataFrame:
        """Download specified sample from MB."""
        return self._make_mb_request({"query": "get_file", "sha256_hash": sha2})

    def get_recent(self: Self, selector: str) -> pd.DataFrame:
        """
        Get the recent MB additions.

        Parameters
        ----------
        selector : str
            Get the latest sample from the last 60 min.

        Returns
        -------
        pd.DataFrame
           The results of the latest addition.

        """
        return self._make_mb_request({"query": "get_recent", "selector": selector})

    def get_cscb(self: Self) -> pd.DataFrame:
        """Query Code Signing Certificate Blocklist (CSCB)."""
        return self._make_mb_request({"query": "get_cscb"})


def _build_query_string(observable: str, mb_type: str, limit: int) -> dict[str, str]:
    """Build the query string for the MB API."""
    query_items: dict[str, Any] = _QUERY_OBJECTS_MAPPINGS[mb_type]
    if "limit" in query_items:
        query_items["limit"] = limit
    for key, value in query_items.items():
        if value == "observable":
            query_items[key] = observable
    return query_items



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Mixin Classes for Sentinel Utilties."""
from __future__ import annotations

import logging
from collections import Counter
from dataclasses import dataclass
from typing import Any

import httpx
import pandas as pd
from azure.common.exceptions import CloudError
from azure.mgmt.core import tools as az_tools
from typing_extensions import Dict, Self, cast

from ..._version import VERSION
from ...auth.azure_auth_core import AzureCloudConfig
from ...common.exceptions import MsticpyAzureConfigError, MsticpyAzureConnectionError
from ...common.pkg_config import get_http_timeout
from .azure_data import AzureData, get_api_headers

__version__ = VERSION
__author__ = "Pete Bryan"

logger: logging.Logger = logging.getLogger(__name__)

_PATH_MAPPING: dict[str, str] = {
    "ops_path": "/providers/Microsoft.SecurityInsights/operations",
    "alert_rules": "/providers/Microsoft.SecurityInsights/alertRules",
    "ss_path": "/savedSearches",
    "bookmarks": "/providers/Microsoft.SecurityInsights/bookmarks",
    "incidents": "/providers/Microsoft.SecurityInsights/incidents",
    "data_connectors": "/providers/Microsoft.SecurityInsights/dataConnectors",
    "watchlists": "/providers/Microsoft.SecurityInsights/watchlists",
    "alert_template": "/providers/Microsoft.SecurityInsights/alertRuleTemplates",
    "search": "/tables",
    "ti": "/providers/Microsoft.SecurityInsights/threatIntelligence/main",
    "dynamic_summary": "/providers/Microsoft.SecurityInsights/dynamicSummaries",
}


@dataclass
class SentinelInstanceDetails:
    """Dataclass for Sentinel workspace details."""

    subscription_id: str
    resource_group: str
    workspace_name: str

    @property
    def resource_id(self) -> str:
        """Return the resource ID for the workspace."""
        return build_sentinel_resource_id(
            self.subscription_id,
            self.resource_group,
            self.workspace_name,
        )

    @classmethod
    def from_resource_id(cls: type[Self], resource_id: str) -> Self:
        """Return SentinelInstanceDetails from a resource ID."""
        return cls(**parse_resource_id(resource_id))


class SentinelUtilsMixin(AzureData):
    """Mixin class for Sentinel core feature integrations."""

    def _get_items(self: Self, url: str, params: dict | None = None) -> httpx.Response:
        """Get items from the API."""
        self.check_connected()
        if params is None:
            params = {"api-version": "2020-01-01"}
        logger.debug("_get_items request to %s.", url)
        if not self._token:
            err_msg = "Token not found, can't get items."
            raise ValueError(err_msg)
        return httpx.get(
            url,
            headers=get_api_headers(self._token),
            params=params,
            timeout=get_http_timeout(),
        )

    def _list_items(  # noqa:PLR0913
        self: Self,
        item_type: str,
        api_version: str = "2020-01-01",
        appendix: str | None = None,
        *,
        next_follow: bool = False,
        params: dict[str, Any] | None = None,
    ) -> pd.DataFrame:
        """
        Return lists of core resources from APIs.

        Parameters
        ----------
        item_type : str
            The type of resource you want to list.
        api_version: str, optional
            The API version to use, by default '2020-01-01'
        appendix: str, optional
            Any appendix that needs adding to the URI, default is None
        next_follow: bool, optional
            If True, follow the nextLink to get all results, by default False
        params: Dict, optional
            Any additional parameters to pass to the API call, by default None

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the requested items.

        Raises
        ------
        CloudError
            If a valid result is not returned.

        """
        item_url: str = (self.url or "") + _PATH_MAPPING[item_type]
        if appendix:
            item_url = item_url + appendix
        if params is None:
            params = {}
        params["api-version"] = api_version
        response: httpx.Response = self._get_items(item_url, params)
        if response.is_success:
            results_df: pd.DataFrame = _azs_api_result_to_df(response)
        else:
            raise CloudError(response=response)
        j_resp: dict[str, Any] = response.json()
        results: list[pd.DataFrame] = [results_df]
        # If nextLink in response, go get that data as well
        if next_follow:
            i = 0
            # Limit to 5 nextLinks to prevent infinite loop
            while "nextLink" in j_resp and i < 5:
                next_url: str = j_resp["nextLink"]
                next_response: httpx.Response = self._get_items(next_url, params)
                next_results_df: pd.DataFrame = _azs_api_result_to_df(next_response)
                results.append(next_results_df)
                j_resp = next_response.json()
                i += 1
        results_df = pd.concat(results)
        logger.info(
            "list_items request to %s returned %d rows",
            item_url,
            len(results_df),
        )
        return results_df

    def _build_sent_res_id(
        self: Self,
        subscription_id: str,
        resource_group: str,
        workspace_name: str,
    ) -> str:
        """
        Build a resource ID.

        Parameters
        ----------
        subscription_id : str
            Subscription ID to use
        resource_group : str
            Resource Group name to use
        workspace_name : str
            Workspace name to user

        Returns
        -------
        str
            The formatted resource ID.

        """
        return build_sentinel_resource_id(
            subscription_id,
            resource_group,
            workspace_name,
        )

    def _build_sentinel_api_root(
        self: Self,
        sentinel_instance: SentinelInstanceDetails,
        base_url: str | None = None,
    ) -> str:
        """
        Build an API URL from an Azure resource ID.

        Parameters
        ----------
        sentinel_instance : SentinelInstanceDetails
            Details of a Sentinel workspace
        base_url : str, optional
            The base URL of the Azure cloud to connect to.
            Defaults to resource manager for configured cloud.
            If no cloud configuration, defaults to resource manager
            endpoint for public cloud.

        Returns
        -------
        str
            A URI to that resource.

        """
        if not base_url:
            base_url = AzureCloudConfig(self.cloud).resource_manager
        resource_id: str | None = sentinel_instance.resource_id
        if base_url.endswith("/"):
            base_url = base_url[:-1]

        sentinel_api_url: str = f"{base_url}{resource_id}"
        logger.info("Sentinel API URL built: %s", sentinel_api_url)
        return sentinel_api_url

    def check_connected(self: Self) -> None:
        """Check that Sentinel workspace is connected."""
        if not self.connected:
            err_msg: str = (
                "Not connected to Sentinel, ensure you run `.connect`"
                "before calling functions."
            )
            raise MsticpyAzureConnectionError(err_msg)


def _azs_api_result_to_df(response: httpx.Response) -> pd.DataFrame:
    """
    Convert API response to a Pandas dataframe.

    Parameters
    ----------
    response : httpx.Response
        A response object from an Azure REST API call.

    Returns
    -------
    pd.DataFrame
        The API response as a Pandas dataframe.

    Raises
    ------
    ValueError
        If the response is not valid JSON.

    """
    j_resp: dict[str, Any] = response.json()
    if not response.is_success or not j_resp:
        err_msg: str = "No valid JSON result in response"
        raise ValueError(err_msg)
    if "value" in j_resp:
        j_resp = j_resp["value"]
    return pd.json_normalize(j_resp)


def build_sentinel_resource_id(
    subscription_id: str,
    resource_group: str,
    workspace_name: str,
) -> str:
    """
    Build a MS Sentinel resource ID.

    Parameters
    ----------
    subscription_id : str
        Subscription ID to use
    resource_group : str
        Resource Group name to use
    workspace_name : str
        Workspace name to user

    Returns
    -------
    str
        The formatted resource ID.

    """
    resource_id: str = (
        f"/subscriptions/{subscription_id}/resourcegroups/{resource_group}"
        f"/providers/Microsoft.OperationalInsights/workspaces/{workspace_name}"
    )
    logger.info("Resource ID built: %s", resource_id)
    return resource_id


def extract_sentinel_response(
    items: dict,
    *,
    props: bool = False,
    etag: dict[str, str] | None = None,
) -> dict:
    """
    Build request data body from items.

    Parameters
    ----------
    items : dict
        A set pf items to be formated in the request body.
    props: bool, optional
        Whether all items are to be built as properities. Default is false.
    etag: dict[str, str], Optional
        If defined, set the etag body value

    Returns
    -------
    dict
        The request body formatted for the API.

    """
    data_body: dict[str, dict[str, str]] = {"properties": {}}
    for key in items:
        if key in ["severity", "status", "title", "message", "searchResults"] or props:
            data_body["properties"].update({key: items[key]})
        else:
            data_body[key] = items[key]
    if etag:
        data_body["etag"] = etag
    return data_body


def validate_resource_id(res_id: str) -> str:
    """Validate a Resource ID String and fix if needed."""
    valid: bool = _validator(res_id)
    if not valid:
        res_id = _fix_res_id(res_id)
        valid = _validator(res_id)
    if not valid:
        err_msg: str = "The Resource ID provided is not valid."
        raise MsticpyAzureConfigError(err_msg)

    return res_id


def parse_resource_id(res_id: str) -> dict[str, Any]:
    """Extract components from workspace resource ID."""
    if not res_id.startswith("/"):
        res_id = f"/{res_id}"
    res_id_parts: Dict[str, str] = cast(
        Dict[str, str], az_tools.parse_resource_id(res_id)
    )
    workspace_name: str | None = None
    if (
        res_id_parts.get("namespace") == "Microsoft.OperationalInsights"
        and res_id_parts.get("type") == "workspaces"
    ):
        workspace_name = res_id_parts.get("name")
    return {
        "subscription_id": res_id_parts.get("subscription"),
        "resource_group": res_id_parts.get("resource_group"),
        "workspace_name": workspace_name,
    }


def _validator(res_id: str) -> bool:
    """Check Resource ID string matches pattern expected."""
    return az_tools.is_valid_resource_id(res_id)


def _fix_res_id(res_id: str) -> str:
    """Try to fix common issues with Resource ID string."""
    if res_id.startswith("https:"):
        res_id = "/".join(res_id.split("/")[5:])
    if not res_id.startswith("/"):
        res_id = "/" + res_id
    if res_id.endswith("/"):
        res_id = res_id[:-1]
    counts = Counter(res_id)
    if counts["/"] > 8:
        res_id = "/".join(res_id.split("/")[:9])
    return res_id



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Base class for Provider classes.

Input can be a single item or a pandas DataFrame containing
multiple items. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from __future__ import annotations

import asyncio
import logging
from abc import ABC, abstractmethod
from asyncio import get_event_loop
from collections.abc import Iterable as C_Iterable
from functools import lru_cache, partial, singledispatch
from typing import TYPE_CHECKING, Any, ClassVar, Coroutine, Generator, Iterable, cast

import pandas as pd
from typing_extensions import Self

from .._version import VERSION
from ..common.utility import export
from ..transform.iocextract import IoCExtract as ItemExtract
from ..transform.iocextract import IoCType
from .lookup_result import LookupStatus, SanitizedObservable
from .preprocess_observable import PreProcessor

if TYPE_CHECKING:
    from ..init.pivot import Pivot
    from ..init.pivot_core.pivot_register import PivotRegistration
    from .lookup import ProgressCounter

__version__ = VERSION
__author__ = "Ian Hellen"

_ITEM_EXTRACT: ItemExtract = ItemExtract()
logger: logging.Logger = logging.getLogger(__name__)


@export
class Provider(ABC):
    """Abstract base class for Providers."""

    _QUERIES: ClassVar[dict[str, Any]] = {}

    @abstractmethod
    def lookup_item(
        self: Self,
        item: str,
        item_type: str | None = None,
        query_type: str | None = None,
    ) -> pd.DataFrame:
        """
        Lookup from a value.

        Parameters
        ----------
        item : str
            item to lookup
        item_type : str, optional
            The Type of the item to lookup, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the item_value
            will be returned.

        Returns
        -------
        pd.DataFrame
            DataFrame of results.

        Raises
        ------
        NotImplementedError
            If attempting to use an HTTP method or authentication
            protocol that is not supported.

        Notes
        -----
        Note: this method uses memoization (lru_cache) to cache results
        for a particular item to try avoid repeated network calls for
        the same item.

        """

    def _check_item_type(
        self: Self,
        item: str,
        item_type: str | None = None,
        query_subtype: str | None = None,
    ) -> dict:
        """
        Check Item Type and cleans up item.

        Parameters
        ----------
        item : str
            item
        item_type : str, optional
            item type, by default None
        query_subtype : str, optional
            Query sub-type, if any, by default None

        Returns
        -------
        Dict
            Dict result with resolved type and pre-processed
            item.
            Status is none-zero on failure.

        """
        item_type = item_type or self.resolve_item_type(item)
        result: dict[str, Any] = {
            "Item": item,
            "SanitizedValue": item,
            "ItemType": item_type,
            "QuerySubtype": query_subtype,
            "Result": False,
            "Details": "",
            "RawResult": None,
            "Reference": None,
            "Status": LookupStatus.OK.value,
        }

        if not self.is_supported_type(result["ItemType"]):
            result["Details"] = f"Type {result['ItemType']} not supported."
            result["Status"] = LookupStatus.NOT_SUPPORTED.value
            return result

        clean_item: SanitizedObservable = self._preprocessors.check(
            item,
            item_type,
            require_url_encoding=self.require_url_encoding,
        )

        result["SanitizedValue"] = clean_item.observable

        if clean_item.status != "ok":
            result["Details"] = clean_item.status
            result["Status"] = LookupStatus.BAD_FORMAT.value

        return result

    # pylint: disable=unused-argument
    def __init__(self: Provider) -> None:
        """Initialize the provider."""
        self.description: str | None = None
        self._supported_types: set[IoCType] = set()
        self._supported_types = {
            IoCType.parse(query_type.split("-")[0]) for query_type in self._QUERIES
        }
        if IoCType.unknown in self._supported_types:
            self._supported_types.remove(IoCType.unknown)

        self.require_url_encoding = False
        self._preprocessors = PreProcessor()

    @property
    def name(self: Self) -> str:
        """Return the name of the provider."""
        return self.__class__.__name__

    def lookup_items(
        self: Self,
        data: pd.DataFrame | dict[str, str] | Iterable[str],
        item_col: str | None = None,
        item_type_col: str | None = None,
        query_type: str | None = None,
    ) -> pd.DataFrame:
        """
        Lookup collection of items.

        Parameters
        ----------
        data : Union[pd.DataFrame, dict[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `item_col` parameter)
            2. Dict of items
            3. Iterable of items
        item_col : str, optional
            DataFrame column to use for items, by default None
        item_type_col : str, optional
            DataFrame column to use for types, by default None
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the type
            will be returned.

        Returns
        -------
        pd.DataFrame
            DataFrame of results.

        """
        results: list[pd.DataFrame] = []
        for item, item_type in generate_items(data, item_col, item_type_col):
            if not item:
                continue
            item_result: pd.DataFrame = self.lookup_item(
                item,
                item_type,
                query_type,
            )
            results.append(item_result)

        if not results:
            return pd.DataFrame()
        return pd.concat(results)

    async def lookup_items_async(  # noqa:PLR0913
        self: Self,
        data: pd.DataFrame | dict[str, str] | Iterable[str],
        item_col: str | None = None,
        item_type_col: str | None = None,
        query_type: str | None = None,
        *,
        prog_counter: ProgressCounter | None = None,
        item_type: str | None = None,
    ) -> pd.DataFrame:
        """
        Lookup collection of items.

        Parameters
        ----------
        data : Union[pd.DataFrame, dict[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `item_col` parameter)
            2. Dict of items, Type
            3. Iterable of items - Types will be inferred
        item_col : str, optional
            DataFrame column to use for items, by default None
        item_type_col : str, optional
            DataFrame column to use for Types, by default None
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the item
            will be returned.
        prog_counter: ProgressCounter, Optional
            Progress Counter to display progess of IOC searches.
        item_type: str, Optional
            Type of item

        Returns
        -------
        pd.DataFrame
            DataFrame of results.

        """
        event_loop: asyncio.AbstractEventLoop = get_event_loop()
        results: list[pd.DataFrame] = []
        for item, ret_item_type in generate_items(data, item_col, item_type_col):
            if not item:
                continue
            get_item = partial(
                self.lookup_item,
                item=item,
                item_type=item_type or ret_item_type,
                query_type=query_type,
            )
            item_result: pd.DataFrame = await event_loop.run_in_executor(None, get_item)
            if prog_counter:
                await prog_counter.decrement()
            results.append(item_result)

        return pd.concat(results)

    @property
    def item_query_defs(self: Self) -> dict[str, Any]:
        """
        Return current dictionary of IoC query/request definitions.

        Returns
        -------
        dict[str, Any]
            IoC query/request definitions keyed by IoCType

        """
        return self._QUERIES

    @classmethod
    def is_known_type(cls: type[Self], item_type: str) -> bool:
        """
        Return True if this a known IoC Type.

        Parameters
        ----------
        item_type : str
            IoCType string to test

        Returns
        -------
        bool
            True if known type.

        """
        return item_type in IoCType.__members__ and item_type != "unknown"

    @property
    def supported_types(self: Self) -> list[str]:
        """
        Return list of supported types for this provider.

        Returns
        -------
        list[str]
            List of supported type names

        """
        return [item.name for item in self._supported_types]

    @classmethod
    def usage(cls: type[Self]) -> None:
        """Print usage of provider."""
        print(f"{cls.__doc__} Supported query types:")
        for key in sorted(cls._QUERIES):
            elements: list[str] = key.split("-", maxsplit=1)
            if len(elements) == 1:
                print(f"\titem_type={elements[0]}")
            if len(elements) > 1:
                print(f"\titem_type={elements[0]}, query_type={elements[1]}")

    def is_supported_type(self: Self, item_type: str | IoCType) -> bool:
        """
        Return True if the passed type is supported.

        Parameters
        ----------
        item_type : Union[str, IoCType]
            type name or instance

        Returns
        -------
        bool
            True if supported.

        """
        if isinstance(item_type, str):
            item_type = IoCType.parse(item_type)
        return item_type.name in self.supported_types

    @staticmethod
    @lru_cache(maxsize=1024)
    def resolve_item_type(item: str) -> str:
        """
        Return IoCType determined by ItemExtract.

        Parameters
        ----------
        item : str
            Item string

        Returns
        -------
        str
            IoCType (or unknown if type could not be determined)

        """
        return _ITEM_EXTRACT.get_ioc_type(item)

    async def _lookup_items_async_wrapper(  # pylint: disable=too-many-arguments # noqa: PLR0913
        self: Self,
        data: pd.DataFrame | dict[str, str] | Iterable[str],
        item_col: str | None = None,
        item_type_col: str | None = None,
        query_type: str | None = None,
        *,
        prog_counter: ProgressCounter | None = None,
    ) -> pd.DataFrame:
        """
        Async wrapper for providers that do not implement lookup_items_async.

        Parameters
        ----------
        data : Union[pd.DataFrame, dict[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `item_col` parameter)
            2. Dict of items, Type
            3. Iterable of items - Type will be inferred
        item_col : str, optional
            DataFrame column to use for items, by default None
        item_type_col : str, optional
            DataFrame column to use for Type, by default None
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoitemC type
            will be returned.
        prog_counter: ProgressCounter; Optional
            Progress Counter to display progress of IOC searches.

        Returns
        -------
        pd.DataFrame
            DataFrame of results.

        """
        event_loop: asyncio.AbstractEventLoop = get_event_loop()
        get_items = partial(
            self.lookup_items,
            data,
            item_col,
            item_type_col,
            query_type,
        )
        result: pd.DataFrame = await event_loop.run_in_executor(None, get_items)
        if prog_counter:
            await prog_counter.decrement(len(data))  # type: ignore[arg-type]
        return result


class PivotProvider(ABC):
    """A class which provides pivot functions and a means of registering them."""

    @abstractmethod
    def register_pivots(
        self: Self,
        pivot_reg: type[PivotRegistration],
        pivot: Pivot,
    ) -> None:
        """
        Register pivot functions for a Provider.

        Parameters
        ----------
        pivot_reg : PivotRegistration
            Pivot registration settings.
        pivot : Pivot
            Pivot library instance

        """


@singledispatch
def generate_items(
    data: pd.DataFrame | dict | C_Iterable,
    item_col: str | None = None,
    item_type_col: str | None = None,
) -> Generator[tuple[str | None, str | None], Any, None]:
    """
    Generate item pairs from different input types.

    Parameters
    ----------
    data : Any
        DataFrame, dictionary or iterable
    item_col : str | None
        If `data` is a DataFrame, the column containing the item value.
    item_type_col : str | None
        If `data` is a DataFrame, the column containing the item type.

    Returns
    -------
    Iterable[tuple[str | None, str | None]] - a tuple of Observable/Type.

    """
    del item_col, item_type_col

    if isinstance(data, C_Iterable):
        for item in data:
            yield cast(str, item), Provider.resolve_item_type(item)
    else:
        yield None, None


@generate_items.register(pd.DataFrame)
def _(
    data: pd.DataFrame,
    item_col: str,
    item_type_col: str | None = None,
) -> Generator[tuple[Any, Any], Any, None]:
    for _, row in data.iterrows():
        if item_type_col is None:
            yield row[item_col], Provider.resolve_item_type(row[item_col])
        else:
            yield row[item_col], row[item_type_col]


@generate_items.register(dict)
def _(
    data: dict,
    _: str | None = None,
    __: str | None = None,
) -> Generator[tuple[Any, Any], Any, None]:
    for item, item_type in data.items():
        if not item_type:
            yield item, Provider.resolve_item_type(item)
        else:
            yield item, item_type


def _make_sync(future: Coroutine) -> pd.DataFrame:
    """Wait for an async call, making it sync."""
    try:
        event_loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
    except RuntimeError:
        # Generate an event loop if there isn't any.
        event_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(event_loop)
    return event_loop.run_until_complete(future)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Local Velociraptor Data Driver class."""
import logging
from collections import defaultdict
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from tqdm.auto import tqdm

from ..._version import VERSION
from ...common.exceptions import MsticpyDataQueryError
from ...common.provider_settings import get_provider_settings
from ...common.utility import export, valid_pyname
from .driver_base import DriverBase, QuerySource

__version__ = VERSION
__author__ = "ianhelle"

logger = logging.getLogger(__name__)

_VELOCIRATOR_DOC_URL = (
    "https://msticpy.readthedocs.io/en/latest/data_acquisition/"
    "DataProv-Velociraptor.html"
)


# pylint: disable=too-many-instance-attributes
@export
class VelociraptorLogDriver(DriverBase):
    """Velociraptor driver class to ingest log data."""

    def __init__(self, connection_str: Optional[str] = None, **kwargs):
        """
        Instantiate Velociraptor driver and optionally connect.

        Parameters
        ----------
        connection_str : str, optional
            Connection string (not used)
        data_paths : List[str], optional
            Paths from which to load data files
        progress : bool, optional
            Show progress with tqdm, by default, True

        """
        del connection_str
        if kwargs.get("debug", False):
            logger.setLevel(logging.DEBUG)
        super().__init__()

        self._paths: List[str] = ["."]
        # If data paths specified, use these
        # from kwargs or settings
        if data_paths := kwargs.get("data_paths"):
            self._paths = [path.strip() for path in data_paths]
            logger.info("data paths read from param %s", str(self._paths))
        else:
            prov_settings = get_provider_settings(config_section="DataProviders").get(
                "VelociraptorLogs"
            )
            if prov_settings:
                self._paths = prov_settings.args.get("data_paths", []) or self._paths
                logger.info("data paths read from config %s", str(self._paths))

        self.data_files: Dict[str, List[Path]] = {}
        self._schema: Dict[str, Any] = {}
        self._query_map: Dict[str, List[str]]
        self._progress = kwargs.pop("progress", True)
        self._loaded = True
        self.has_driver_queries = True
        logger.info("data files to read %s", ",".join(self.data_files))

    def connect(self, connection_str: Optional[str] = None, **kwargs):
        """
        Connect to data source.

        Parameters
        ----------
        connection_str : str
            Connect to a data source

        """
        del connection_str
        self.data_files = self._get_logfile_paths()
        self._connected = True

    @property
    def schema(self) -> Dict[str, Dict]:
        """
        Return current data schema of connection.

        Returns
        -------
        Dict[str, Dict]
            Data schema of current connection.

        """
        if not self._schema:
            if not self.data_files:
                self.connect()
            # read the first row of each file to get the schema
            iter_data_files = (
                tqdm(self.data_files.items())
                if self._progress
                else self.data_files.items()
            )
            for table, files in iter_data_files:
                if not files:
                    continue
                sample_df = pd.read_json(files[0], lines=True, nrows=1)
                # pylint: disable=no-member
                self._schema[table] = {
                    col: dtype.name for col, dtype in sample_df.dtypes.to_dict().items()
                }
            logger.info("Reading schema for %d tables", len(self.data_files))

        return self._schema

    def query(
        self, query: str, query_source: Optional[QuerySource] = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Returns
        -------
        Union[pd.DataFrame, results.ResultSet]
            A DataFrame (if successful) or
            the underlying provider result if an error.

        """
        del kwargs
        if not self.data_files:
            self.connect()

        if query not in self.data_files:
            raise MsticpyDataQueryError(
                f"No data loaded for query {query}.",
                "Please check that the data loaded from the log paths",
                "has this data type.",
            )
        return self._cached_query(query)

    @lru_cache(maxsize=256)
    def _cached_query(self, query: str) -> pd.DataFrame:
        iter_data_files = (
            tqdm(self.data_files[query]) if self._progress else self.data_files[query]
        )
        dfs = [pd.read_json(file, lines=True) for file in iter_data_files]
        query_df = pd.concat(dfs, ignore_index=True)

        logger.info("Query %s, returned %d rows", query, len(query_df))
        return query_df

    def query_with_results(self, query, **kwargs):
        """Return query with fake results."""
        return self.query(query, **kwargs), "OK"

    @property
    def driver_queries(self) -> List[Dict[str, Any]]:
        """
        Return dynamic queries available on connection to data.

        Returns
        -------
        List[Dict[str, Any]]
            List of queries with properties: "name", "query", "container"
            and (optionally) "description"

        Raises
        ------
        MsticpyNotConnectedError
            If called before driver is connected.

        """
        if not self.connected:
            self.connect()
        if self.data_files:
            return [
                {
                    "name": query,
                    "query": query,
                    "query_paths": "velociraptor",
                    "description": f"Velociraptor {query} table.",
                }
                for query in self.data_files
            ]
        return []

    def _get_logfile_paths(self) -> Dict[str, List[Path]]:
        """Read files in data paths."""
        data_files: Dict[str, List[Path]] = defaultdict(list)

        for input_path in (Path(path_str) for path_str in self._paths):
            files = {
                file.relative_to(input_path): file
                for file in input_path.rglob("*.json")
            }

            file_names = [valid_pyname(str(file.with_suffix(""))) for file in files]
            path_files = dict(zip(file_names, files.values()))
            for file_name, file_path in path_files.items():
                data_files[file_name].append(file_path)

        logger.info("Found %d data file types", len(data_files))
        logger.info("Total data files: %d", sum(len(v) for v in data_files.values()))
        if not data_files:
            raise MsticpyDataQueryError(
                "No usable data files found in supplied paths.",
                f"Data paths supplied: {', '.join(self._paths)}",
                title="No data files found",
                help_uri=_VELOCIRATOR_DOC_URL,
            )
        return data_files



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Sentinel Alert class."""
import json
from typing import Any, Dict, List

import pandas as pd

from ...datamodel.entities import Alert

_FEATURES = [
    "AdditionalData",
    "AlertLink",
    "AlertName",
    "AlertSeverity",
    "ProcessingEndTime",
    "ResourceId",
    "SourceComputerId",
    "SourceSystem",
    "Status",
    "Tactics",
    "TenantId",
    "VendorOriginalId",
]

_ID_PROPERTIES: List[str] = [
    "AzSubscriptionId",
    "AzResourceId",
    "WorkspaceId",
    "AgentId",
    "TenantId",
    "SourceComputerId",
    "ResourceId",
    "WorkspaceSubscriptionId",
    "WorkspaceResourceGroup",
    "ProviderAlertId",
    "SystemAlertId",
    "ResourceIdentifiers",
]

_QUERY_PROPERTIES: List[str] = [
    "Query Period",
    "Trigger Operator",
    "Trigger Threshold",
    "Search Query Results Overall Count",
    "Data Sources",
    "Query",
    "Query Start Time UTC",
    "Query End Time UTC",
    "Analytic Rule Ids",
    "Event Grouping",
    "Analytic Rule Name",
]


class SentinelAlert(Alert):
    """Security Alert Class."""

    def __init__(
        self,
        src_event=None,
        src_entity=None,
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties such as a Pandas Series
            (the default is None)

        """
        self._custom_query_params: Dict[str, Any] = {}
        super().__init__(src_entity, src_event, **kwargs)
        if (
            isinstance(
                src_event,
                pd.Series,
            )
            and not src_event.empty
        ):
            self._add_sentinel_items(src_event)
            self._add_extended_sent_props()
        self._ids: Dict[str, str] = {}
        if self.__dict__ is not None:
            for id_property in _ID_PROPERTIES:
                if id_property in self.properties:
                    self._ids[id_property] = self.properties[id_property]

    def _add_extended_sent_props(self):
        if self.ExtendedProperties:
            try:
                ext_ents = json.loads(self.ExtendedProperties)
                for item_name, value in ext_ents.items():
                    self.__dict__[item_name] = value
            except json.JSONDecodeError:
                pass
        if self.AdditionalData:
            for item_name, value in self.AdditionalData.items():
                if item_name not in self.__dict__:
                    self.__dict__[item_name] = value

    def _add_sentinel_items(self, src_event):
        for feature in _FEATURES:
            if feature not in self.__dict__:
                self.__dict__.update({feature: src_event.get(feature, "")})

    @property
    def ids(self) -> Dict[str, str]:
        """Return a collection of Identity properties for the alert."""
        return self._ids

    @property
    def is_in_workspace(self) -> bool:
        """Return True if the alert has a Log Analytics WorkspaceID."""
        return "WorkspaceId" in self._ids and "AgentId" in self._ids

    @property
    def is_in_log_analytics(self) -> bool:
        """Return True if the alert originates from a Log Analytics Workspace host."""
        return "TenantId" in self._ids

    @property
    def links(self):
        """Return all links associated with the alert."""
        alert_links = {}
        if "ExtendedLinks" in self.__dict__:
            try:
                links = json.loads(self.__dict__["ExtendedLinks"])
                for link in links:
                    alert_links[link["Label"]] = link["Href"]
            except json.JSONDecodeError:
                pass
        return alert_links

    @property
    def incident(self):
        """Return any incident details associated with the alert."""
        if "InvestigationName" not in self.__dict__:
            self.__dict__["InvestigationName"] = None
        return self.InvestigationName

    @property
    def analytic(self):
        """Return any Sentinel Analytics associated with the alert."""
        return {
            query_property: self.__dict__[query_property]
            for query_property in _QUERY_PROPERTIES
            if query_property in self.__dict__
        }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
ip_utils - IP Address functions.

Contains a series of functions required to manipulate and
enrich IP Address data to assist investigations.

Designed to support any data source containing IP address entity.

"""
from __future__ import annotations

import ipaddress
import logging
import re
import socket
import warnings
from dataclasses import asdict, dataclass, field
from functools import lru_cache
from time import sleep
from typing import Any, Callable, Iterator

import httpx
import pandas as pd
from bs4 import BeautifulSoup
from deprecated.sphinx import deprecated
from typing_extensions import Self

from .._version import VERSION
from ..common.exceptions import MsticpyConnectionError, MsticpyException
from ..common.utility import arg_to_list, export
from ..datamodel.entities import GeoLocation, IpAddress

logger: logging.Logger = logging.getLogger(__name__)

__version__ = VERSION
__author__ = "Ashwin Patil"


_REGISTRIES: dict[str, dict[str, str]] = {
    "arin": {
        "url": "http://rdap.arin.net/registry/ip/",
    },
    "ripencc": {
        "url": "http://rdap.db.ripe.net/ip/",
    },
    "apnic": {
        "url": "http://rdap.apnic.net/ip/",
    },
    "lacnic": {
        "url": "http://rdap.lacnic.net/rdap/ip/",
    },
    "afrinic": {
        "url": "http://rdap.afrinic.net/rdap/ip/",
    },
}

_POTAROO_ASNS_URL = "https://bgp.potaroo.net/cidr/autnums.html"

RATE_LIMIT_THRESHOLD: int = 50


# Closure to cache ASN dictionary from Potaroo
def _fetch_asns() -> Callable[[], dict[str, str]]:
    """Create closure for ASN fetching."""
    asns_dict: dict[str, str] = {}

    def _get_asns_dict() -> dict[str, str]:
        """Return or fetch and return ASN Soup."""
        nonlocal asns_dict
        if not asns_dict:
            try:
                asns_resp: httpx.Response = httpx.get(_POTAROO_ASNS_URL)
            except httpx.ConnectError as err:
                err_msg: str = "Unable to get ASN details from potaroo.net"
                raise MsticpyConnectionError(err_msg) from err
            asns_soup = BeautifulSoup(asns_resp.content, features="lxml")
            asns_dict = {
                str(asn.next_element)
                .strip(): str(asn.next_element.next_element)
                .strip()
                for asn in asns_soup.find_all("a")
            }
        return asns_dict

    return _get_asns_dict


# Create the dictionary accessor from the fetch_asns wrapper
_ASNS_DICT: Callable[[], dict[str, str]] = _fetch_asns()


@export
def convert_to_ip_entities(
    ip_str: str | None = None,
    data: pd.DataFrame | None = None,
    ip_col: str | None = None,
    *,
    geo_lookup: bool = True,
) -> list[IpAddress]:
    """
    Take in an IP Address string and converts it to an IP Entity.

    Parameters
    ----------
    ip_str : str
        A string with a single IP Address or multiple addresses
        delimited by comma or space
    data : pd.DataFrame
        Use DataFrame as input
    ip_col : str
        Column containing IP addresses
    geo_lookup : bool
        If true, do geolocation lookup on IPs,
        by default, True

    Returns
    -------
    List
        The populated IP entities including address and geo-location

    Raises
    ------
    ValueError
        If neither ip_string or data/column provided as input

    """
    # locally imported to prevent cyclic import
    # pylint: disable=import-outside-toplevel, cyclic-import
    from .geoip import GeoLiteLookup

    geo_lite_lookup: GeoLiteLookup = GeoLiteLookup()

    ip_entities: list[IpAddress] = []
    all_ips: set[str] = set()

    if ip_str:
        addrs: list[str] = arg_to_list(ip_str)
    elif data is not None and ip_col:
        addrs = data[ip_col].to_numpy().tolist()
    else:
        err_msg: str = "Must specify either ip_str or data + ip_col parameters."
        raise ValueError(err_msg)

    for addr in addrs:
        if isinstance(addr, list):
            ip_list: set[str] = set(addr)
        elif isinstance(addr, str) and "," in addr:
            ip_list = {ip.strip() for ip in addr.split(",")}
        else:
            ip_list = {addr}
        ip_list = ip_list - all_ips  # remove IP addresses we've seen
        ip_entities.extend(IpAddress(Address=ip) for ip in ip_list)
        all_ips |= ip_list
        if geo_lookup:
            for ip_ent in ip_entities:
                geo_lite_lookup.lookup_ip(ip_entity=ip_ent)
    return ip_entities


@deprecated("Will be removed in a version 2.2", version="1.4.0")
@export
def create_ip_record(
    heartbeat_df: pd.DataFrame,
    az_net_df: pd.DataFrame | None = None,
) -> IpAddress:
    """
    Generate ip_entity record for provided IP value.

    Parameters
    ----------
    heartbeat_df : pd.DataFrame
        A dataframe of heartbeat data for the host
    az_net_df : pd.DataFrame
        Option dataframe of Azure network data for the host

    Returns
    -------
    IP
        Details of the IP data collected

    """
    ip_entity: IpAddress = IpAddress()

    # Produce ip_entity record using available dataframes
    ip_hb: pd.Series[str] = heartbeat_df.iloc[0]
    ip_entity.Address = ip_hb["ComputerIP"]
    ip_entity.hostname = ip_hb["Computer"]
    ip_entity.SourceComputerId = ip_hb["SourceComputerId"]
    ip_entity.OSType = ip_hb["OSType"]
    ip_entity.OSName = ip_hb["OSName"]
    ip_entity.OSVMajorVersion = ip_hb["OSMajorVersion"]
    ip_entity.OSVMinorVersion = ip_hb["OSMinorVersion"]
    ip_entity.ComputerEnvironment = ip_hb["ComputerEnvironment"]
    ip_entity.OmsSolutions = [sol.strip() for sol in ip_hb["Solutions"].split(",")]
    ip_entity.VMUUID = ip_hb["VMUUID"]
    ip_entity.SubscriptionId = ip_hb["SubscriptionId"]
    geoloc_entity: GeoLocation = GeoLocation()
    geoloc_entity.CountryOrRegionName = ip_hb["RemoteIPCountry"]
    try:
        geoloc_entity.Longitude = float(ip_hb["RemoteIPLongitude"])
        geoloc_entity.Latitude = float(ip_hb["RemoteIPLatitude"])
    except TypeError:
        pass
    ip_entity.Location = geoloc_entity

    # If Azure network data present add this to host record
    if az_net_df is not None and not az_net_df.empty:
        if len(az_net_df) == 1:
            priv_addr_str: str = az_net_df["PrivateIPAddresses"].loc[0]
            ip_entity["private_ips"] = convert_to_ip_entities(priv_addr_str)
            pub_addr_str: str = az_net_df["PublicIPAddresses"].loc[0]
            ip_entity["public_ips"] = convert_to_ip_entities(pub_addr_str)
        else:
            if "private_ips" not in ip_entity:
                ip_entity["private_ips"] = []
            if "public_ips" not in ip_entity:
                ip_entity["public_ips"] = []

    return ip_entity


@export
def get_ip_type(ip: str | None = None, ip_str: str | None = None) -> str:
    """
    Validate value is an IP address and determine IPType category.

    (IPAddress category is e.g. Private/Public/Multicast).

    Parameters
    ----------
    ip : str
        The string of the IP Address
    ip_str : str
        The string of the IP Address - alias for `ip`

    Returns
    -------
    str
        Returns ip type string using ip address module

    """
    ip_str = ip or ip_str
    if not ip_str:
        err_msg: str = "'ip' or 'ip_str' value must be specified"
        raise ValueError(err_msg)
    try:
        ip_obj: ipaddress.IPv4Address | ipaddress.IPv6Address = ipaddress.ip_address(
            ip_str,
        )
    except ValueError:
        logger.exception("%s does not appear to be an IPv4 or IPv6 address", ip_str)
    else:
        return_values: dict[str, str] = {
            "is_multicast": "Multicast",
            "is_global": "Public",
            "is_loopback": "Loopback",
            "is_link_local": "Link Local",
            "is_unspecified": "Unspecified",
            "is_private": "Private",
            "is_reserved": "Reserved",
        }
        for func, msg in return_values.items():
            if getattr(ip_obj, func):
                return msg
        return "Unspecified"

    return "Unspecified"


@deprecated("Replaced with ip_whois function", version="2.1.0")
@export
@lru_cache(maxsize=1024)
def get_whois_info(
    ip: str | None = None,
    *,
    show_progress: bool = False,
    ip_str: str | None = None,
) -> pd.DataFrame | _IpWhoIsResult:
    """
    Retrieve whois ASN information for given IP address using IPWhois python package.

    Parameters
    ----------
    ip : str
        IP Address to look up.
    ip_str : str
        alias for `ip`.
    show_progress : bool, optional
        Show progress for each query, by default False

    Returns
    -------
    IP
        Details of the IP data collected

    Notes
    -----
    This function uses the Python functools lru_cache and
    will return answers from the cache for previously queried
    IP addresses.

    """
    ip_str = ip or ip_str
    if not ip_str:
        err_msg: str = "'ip' or 'ip_str' value must be specified"
        raise ValueError(err_msg)
    ip_type: str = get_ip_type(ip_str)
    if ip_type == "Public":
        logger.info(ip_str)
        try:
            whois_result: pd.DataFrame | _IpWhoIsResult = ip_whois(ip_str)
        except MsticpyException as err:
            return _IpWhoIsResult(
                name=f"Error during lookup of {ip_str} {type(err)}",
                properties={},
            )
        if show_progress:
            logger.info(".")
        return whois_result
    return _IpWhoIsResult(
        name=f"No ASN Information for IP type: {ip_type}",
        properties={},
    )


@export
def get_whois_df(  # noqa: PLR0913
    data: pd.DataFrame,
    ip_column: str,
    *,
    all_columns: bool = True,
    asn_col: str = "AsnDescription",
    whois_col: str = "WhoIsData",
    show_progress: bool = False,
) -> pd.DataFrame:
    """
    Retrieve Whois ASN information for DataFrame of IP Addresses.

    Parameters
    ----------
    data : pd.DataFrame
        Input DataFrame
    ip_column : str
        Column name of IP Address to look up.
    all_columns:
        Expand all whois data to columns.
    asn_col : str, optional
        Name of the output column for ASN description,
        by default "ASNDescription".
        Ignored if `all_columns` is True.
    whois_col : str, optional
        Name of the output column for full whois data,
        by default "WhoIsData"
        Ignored if `all_columns` is True.
    show_progress : bool, optional
        Show progress for each query, by default False

    Returns
    -------
    pd.DataFrame
        Output DataFrame with results in added columns.

    """
    del show_progress
    whois_data: pd.DataFrame | _IpWhoIsResult = ip_whois(
        data[ip_column].drop_duplicates(),
    )
    if (
        isinstance(whois_data, pd.DataFrame)
        and not whois_data.empty
        and "query" in whois_data.columns
    ):
        data = data.merge(
            whois_data,
            how="left",
            left_on=ip_column,
            right_on="query",
            suffixes=("", "_whois"),
        )
        data[whois_col] = data[whois_data.columns].apply(lambda x: x.to_dict(), axis=1)
        data[asn_col] = data["asn_description"]
        if not all_columns:
            return data.drop(columns=whois_data.columns)
        return data
    return data.assign(ASNDescription="No data returned")


@pd.api.extensions.register_dataframe_accessor("mp_whois")
@export
class IpWhoisAccessor:
    """Pandas api extension for IP Whois lookup."""

    def __init__(self: IpWhoisAccessor, pandas_obj: pd.DataFrame) -> None:
        """Instantiate pandas extension class."""
        self._df: pd.DataFrame = pandas_obj

    def lookup(
        self: Self,
        ip_column: str,
        *,
        asn_col: str = "ASNDescription",
        whois_col: str = "WhoIsData",
        show_progress: bool = False,
    ) -> pd.DataFrame:
        """
        Extract IoCs from either a pandas DataFrame.

        Parameters
        ----------
        ip_column : str
            Column name of IP Address to look up.
        asn_col : str, optional
            Name of the output column for ASN description,
            by default "ASNDescription"
        whois_col : str, optional
            Name of the output column for full whois data,
            by default "WhoIsData"
        show_progress : bool, optional
            Show progress for each query, by default False

        Returns
        -------
        pd.DataFrame
            Output DataFrame with results in added columns.

        """
        warn_message = (
            "This accessor method has been deprecated.\n"
            "Please use IpAddress.util.whois() pivot function."
            "This will be removed in MSTICPy v2.2.0"
        )
        warnings.warn(
            warn_message,
            category=DeprecationWarning,
            stacklevel=1,
        )
        return get_whois_df(
            data=self._df,
            ip_column=ip_column,
            asn_col=asn_col,
            whois_col=whois_col,
            show_progress=show_progress,
        )


def ip_whois(
    ip: IpAddress | str | list | pd.Series | None = None,
    ip_address: IpAddress | str | list[str] | pd.Series | None = None,
    *,
    raw: bool = False,
    query_rate: float = 0.5,
    retry_count: int = 5,
) -> pd.DataFrame | _IpWhoIsResult:
    """
    Lookup IP Whois information.

    Parameters
    ----------
    ip : Union[IpAddress, str, List]
        An IP address or list of IP addresses to lookup.
    ip_address : Union[IpAddress, str, List]
        An IP address or list of IP addresses to lookup. Alias of `ip`.
    raw : bool, optional
        Set True if raw whois result wanted, by default False
    query_rate : float, optional
        Controls the rate at which queries are made, by default 0.5
    retry_count : int, optional
        The number of times to retry a query if it fails, default is 5

    Returns
    -------
    Union[pd.DataFrame, Dict]
        If a single IP Address provided result is a dictionary
        if multiple provided results are formatted as a dataframe.

    Raises
    ------
    ValueError:
        If neither `ip` nor `ip_address` is supplied.

    """
    ip = ip if ip is not None else ip_address
    if ip is None:
        err_msg: str = "One of ip or ip_address parameters must be supplied."
        raise ValueError(err_msg)
    if isinstance(ip, (list, pd.Series)):
        rate_limit: bool = len(ip) > RATE_LIMIT_THRESHOLD
        if rate_limit:
            logger.info("Large number of lookups, this may take some time.")
        whois_results: dict[str, Any] = {}
        for ip_addr in ip:
            if rate_limit:
                sleep(query_rate)
            whois_results[ip_addr] = _whois_lookup(  # type: ignore[index]
                ip_addr,
                raw=raw,
                retry_count=retry_count,
            ).properties
        return _whois_result_to_pandas(whois_results)
    if isinstance(ip, (str, IpAddress)):
        return _whois_lookup(ip, raw=raw)
    return pd.DataFrame()


def get_asn_details(asns: str | list[str]) -> pd.DataFrame | dict[str, Any]:
    """
    Get details about an ASN(s) from its number.

    Parameters
    ----------
    asns : Union[str, List]
        A single ASN or list of ASNs to lookup.

    Returns
    -------
    Union[pd.DataFrame, Dict]
        Details about the ASN(s) if a single ASN provided, result is a dictionary.
        If multiple provided, results are returned as a DataFrame.

    """
    if isinstance(asns, list):
        asn_detail_results: list[dict[str, Any]] = [
            _asn_results(str(asn)) for asn in asns
        ]
        return pd.DataFrame(asn_detail_results)
    return _asn_results(str(asns))


def get_asn_from_name(name: str) -> dict[str, Any]:
    """
    Get a list of ASNs that match a name.

    Parameters
    ----------
    name : str
        The name of the ASN to search for

    Returns
    -------
    Dict
        A list of ASNs that match the name.

    Raises
    ------
    MsticpyConnectionError
        If the list of all ASNs cannot be retrieved.
    MsticpyException
        If no matches found.

    """
    name = name.casefold()
    asns_dict: dict[str, str] = _ASNS_DICT()
    matches: dict[str, str] = {
        key: value for key, value in asns_dict.items() if name in value.casefold()
    }
    if len(matches.keys()) == 1:
        return next(iter(matches))  # type:ignore[arg-type]
    if len(matches.keys()) > 1:
        return matches

    err_msg: str = f"No ASNs found matching {name}"
    raise MsticpyException(err_msg)


def get_asn_from_ip(
    ip: str | IpAddress | None = None,
    ip_address: str | IpAddress | None = None,
) -> dict[str, Any]:
    """
    Get the ASN that an IP belongs to.

    Parameters
    ----------
    ip : Union[str, IpAddress]
        IP address to lookup.
    ip_address : Union[str, IpAddress]
        IP address to lookup (alias of `ip`).

    Returns
    -------
    dict
        Details of the ASN that the IP belongs to.

    """
    ip_param: str | IpAddress | None = ip or ip_address
    if not ip_param:
        return {}
    if isinstance(ip_param, IpAddress):
        ip_param = ip_param.Address
    ip_str: str = ip_param.strip()
    query: str = f" -v {ip_str}\r\n"
    ip_response: str = _cymru_query(query)
    keys: list[str] = ip_response.split("\n", maxsplit=1)[0].split("|")
    values: list[str] = ip_response.split("\n")[1].split("|")
    return {key.strip(): value.strip() for key, value in zip(keys, values)}


@dataclass
class _IpWhoIsResult:
    """Named tuple for IPWhoIs Result."""

    name: str | None = None
    properties: dict[str, Any] = field(default_factory=dict)

    # Backward compatibility with namedtuple
    def __iter__(self) -> Iterator[Any]:
        """Iterate over properties."""
        return iter(asdict(self).values())

    def __getitem__(self, item):
        """Get item from properties."""
        return list(asdict(self).values())[item]


@lru_cache(maxsize=1024)
def _whois_lookup(
    ip_addr: str | IpAddress,
    *,
    raw: bool = False,
    retry_count: int = 5,
) -> _IpWhoIsResult:
    """Conduct lookup of IP Whois information."""
    if isinstance(ip_addr, IpAddress):
        ip_addr = ip_addr.Address
    asn_items: dict[str, Any] = get_asn_from_ip(ip_addr.strip())
    registry_url: str | None = None
    if asn_items and "Error: no ASN or IP match on line 1." not in asn_items:
        ipwhois_result: _IpWhoIsResult = _IpWhoIsResult(asn_items["AS Name"], {})
        ipwhois_result.properties["asn"] = asn_items["AS"]
        ipwhois_result.properties["query"] = asn_items["IP"]
        ipwhois_result.properties["asn_cidr"] = asn_items["BGP Prefix"]
        ipwhois_result.properties["asn_country_code"] = asn_items["CC"]
        ipwhois_result.properties["asn_registry"] = asn_items["Registry"]
        ipwhois_result.properties["asn_date"] = asn_items["Allocated"]
        ipwhois_result.properties["asn_description"] = asn_items["AS Name"]
        registry_url = _REGISTRIES.get(asn_items.get("Registry", ""), {}).get("url")
    if not asn_items or not registry_url:
        return _IpWhoIsResult(None)
    return _add_rdap_data(
        ipwhois_result=ipwhois_result,
        rdap_reg_url=f"{registry_url}{ip_addr}",
        retry_count=retry_count,
        raw=raw,
    )


def _add_rdap_data(
    ipwhois_result: _IpWhoIsResult,
    rdap_reg_url: str,
    retry_count: int,
    *,
    raw: bool,
) -> _IpWhoIsResult:
    """Add RDAP data to WhoIs result."""
    retries = 0
    while retries < retry_count:
        rdap_data: httpx.Response = _rdap_lookup(
            url=rdap_reg_url,
            retry_count=retry_count,
        )
        if rdap_data.is_success:
            rdap_data_content: dict[str, Any] = rdap_data.json()
            net: dict[str, Any] = _create_net(rdap_data_content)
            ipwhois_result.properties["nets"] = [net]
            for link in rdap_data_content["links"]:
                if link["rel"] == "up":
                    up_data_link: str = link["href"]
                    up_rdap_data: httpx.Response = httpx.get(up_data_link)
                    up_rdap_data_content: dict[str, Any] = up_rdap_data.json()
                    up_net: dict[str, Any] = _create_net(up_rdap_data_content)
                    ipwhois_result.properties["nets"].append(up_net)
            if raw:
                ipwhois_result.properties["raw"] = rdap_data
            break
        if rdap_data.status_code == httpx.codes.TOO_MANY_REQUESTS:
            sleep(3)
            retries += 1
            continue
        err_msg: str = f"Error code: {rdap_data.status_code}"
        raise MsticpyConnectionError(err_msg)
    return ipwhois_result


def _rdap_lookup(url: str, retry_count: int = 5) -> httpx.Response:
    """Perform RDAP lookup with retries."""
    rdap_data: httpx.Response | None = None
    while retry_count > 0 and not rdap_data:
        rdap_data = _run_rdap_query(url)
        retry_count -= 1
    if not rdap_data:
        err_msg: str = (
            "Rate limit exceeded - try adjusting query_rate parameter "
            "to slow down requests"
        )
        raise MsticpyException(err_msg)
    return rdap_data


def _run_rdap_query(url: str) -> httpx.Response | None:
    """Execute rdap query call and handle errors."""
    try:
        return httpx.get(url)
    except (httpx.WriteError, httpx.ReadError):
        return None


def _whois_result_to_pandas(results: str | list[str] | dict[str, Any]) -> pd.DataFrame:
    """Transform whois results to a Pandas DataFrame."""
    if isinstance(results, dict):
        return pd.DataFrame(
            [result or {"query": idx} for idx, result in results.items()],
        )
    err_msg: str = "Only dict type current supported for `results`."
    raise NotImplementedError(err_msg)


def _find_address(
    entity: dict,
) -> str | None:
    """Find an orgs address from an RDAP entity."""
    if "vcardArray" not in entity:
        return None
    for vcard in [vcard for vcard in entity["vcardArray"] if isinstance(vcard, list)]:
        for vcard_sub in vcard:
            if vcard_sub[0] == "adr" and "label" in vcard_sub[1]:
                return vcard_sub[1]["label"]
    return None


def _create_net(data: dict[str, Any]) -> dict[str, Any]:
    """Create a network object from RDAP data."""
    net_data: dict[str, Any] = data.get("cidr0_cidrs", [None])[0] or {}
    net_prefixes: set[str] = net_data.keys() & {"v4prefix", "v6prefix"}

    if not net_data or not net_prefixes:
        net_cidr: str = "No network data retrieved."
    else:
        net_cidr = " ".join(
            f"{net_data[net_prefix]}/{net_data.get('length', '')}"
            for net_prefix in net_prefixes
        )
    address: str | None = None
    created: str | None = None
    updated: str | None = None
    for item in data["events"]:
        created = item["eventDate"] if item["eventAction"] == "last changed" else None
        updated = item["eventDate"] if item["eventAction"] == "registration" else None
    for entity in data["entities"]:
        address = _find_address(entity)
    regex = r"[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+@[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*"
    emails: list[str] = re.findall(regex, str(data))
    return {
        "cidr": net_cidr,
        "handle": data["handle"],
        "name": data["name"],
        "startAddress": data["startAddress"],
        "endAddress": data["endAddress"],
        "created": created,
        "updated": updated,
        "address": address,
        "emails": emails,
    }


def _asn_whois_query(
    query: str,
    server: str,
    port: int = 43,
    retry_count: int = 5,
) -> str:
    """Connect to whois server and send query."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as conn:
        conn.connect((server, port))
        conn.send(query.encode())
        response: list[str] = []
        response_data: str | None = None
        while retry_count > 0 and not response_data:
            response_data = _run_asn_query(conn, response)
            retry_count -= 1
    return "".join(response)


def _run_asn_query(
    conn: socket.socket,
    response: list[str],
) -> str | None:
    """Execute asn query call and handle errors."""
    try:
        response_data: str = conn.recv(4096).decode()
    except (UnicodeDecodeError, ConnectionResetError):
        return None
    if "error" in response_data:
        err_msg: str = "An error occurred during lookup, please try again."
        raise MsticpyConnectionError(err_msg)
    if "rate limit exceeded" in response_data:
        err_msg = "Rate limit exceeded please wait and try again."
        raise MsticpyConnectionError(err_msg)
    response.append(response_data)
    return response_data


def _cymru_query(query: str) -> str:
    """Query Cymru for ASN information."""
    return _asn_whois_query(query, "whois.cymru.com")


def _radb_query(query: str) -> str:
    """Query RADB for ASN information."""
    return _asn_whois_query(query, "whois.radb.net")


def _parse_asn_response(response: str) -> dict[str, Any]:
    """Parse ASN response into a dictionary."""
    response_output: dict[str, Any] = {}
    for item in response.split("\n"):
        try:
            key: str = item.split(":")[0].strip()
            if key and key not in response_output:
                try:
                    response_output[key] = item.split(":")[1].strip()
                except IndexError:
                    continue
            elif key in response_output:
                if not isinstance(response_output[key], list):
                    response_output[key] = [response_output[key]]
                response_output[key].append(item.split(":")[1].strip())
        except IndexError:
            continue
    return response_output


def _asn_results(asn: str) -> dict[str, Any]:
    """Get ASN details from ASN number."""
    if not asn.startswith("AS"):
        asn = f"AS{asn}"
    query1: str = f" {asn}\r\n"
    asn_response: str = _radb_query(query1)
    asn_details: dict[str, Any] = _parse_asn_details(asn_response)
    query2: str = f" -i origin {asn}\r\n"
    asn_ranges_response: str = _radb_query(query2)
    asn_details["ranges"] = _parse_asn_ranges(asn_ranges_response)
    return asn_details


def _parse_asn_details(response: str) -> dict[str, Any]:
    """Parse ASN details response into a dictionary."""
    asn_keys: list[str] = ["aut-num", "as-name", "descr", "notify", "changed"]
    asn_output: dict[str, Any] = _parse_asn_response(response)
    asn_output_filtered: dict[str, Any] = {
        key: value for key, value in asn_output.items() if key in asn_keys
    }
    asn_output_filtered["Autonomous Number"] = asn_output_filtered.pop("aut-num", None)
    asn_output_filtered["AS Name"] = asn_output_filtered.pop("as-name", None)
    asn_output_filtered["Description"] = asn_output_filtered.pop("descr", None)
    asn_output_filtered["Contact"] = asn_output_filtered.pop("notify", None)
    asn_output_filtered["Last Updated"] = asn_output_filtered.pop("changed", None)
    return asn_output_filtered


def _parse_asn_ranges(response: str) -> list[str]:
    """Parse ASN ranges response into a list."""
    return [
        item.split(":   ")[1].strip()
        for item in response.split("\n")
        if item.split(":   ")[0].strip() == "route"
    ]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Generic Module for Lookup classes.

Input can be a single item or a pandas DataFrame containing
multiple items. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from __future__ import annotations

import asyncio
import importlib
import logging
import warnings
from collections import ChainMap
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Iterable, Mapping, Sized

import nest_asyncio
import pandas as pd
from tqdm.auto import tqdm
from typing_extensions import Self

from .._version import VERSION
from ..common.exceptions import MsticpyConfigError, MsticpyUserConfigError
from ..common.provider_settings import (
    ProviderSettings,
    get_provider_settings,
    reload_settings,
)
from ..common.utility import export, is_ipython
from ..vis.ti_browser import browse_results
from .lookup_result import LookupStatus

# used in dynamic instantiation of providers
from .provider_base import Provider, _make_sync

if TYPE_CHECKING:
    import datetime
    from types import ModuleType

    from ..nbwidgets.select_item import SelectItem

__version__ = VERSION
__author__ = "Florian Bracq"

logger: logging.Logger = logging.getLogger(__name__)

_HTTP_PROVIDER_LEGAL_KWARGS: list[str] = ["timeout", "ApiID", "AuthKey", "Instance"]


class ProgressCounter:
    """Progress counter for async tasks."""

    def __init__(self: ProgressCounter, total: int) -> None:
        """Initialize the class."""
        self.total: int = total
        self._lock: asyncio.Condition = asyncio.Condition()
        self._remaining: int = total

    async def decrement(self: Self, increment: int = 1) -> None:
        """Decrement the counter."""
        if self._remaining == 0:
            return
        async with self._lock:
            self._remaining -= increment

    async def get_remaining(self: Self) -> int:
        """Get the current remaining count."""
        async with self._lock:
            return self._remaining


@export
class Lookup:
    """Item lookup from providers."""

    _NO_PROVIDERS_MSG: ClassVar[
        str
    ] = """
    No Providers are loaded - please check that
    you have correctly configured your msticpyconfig.yaml settings.
    """

    _HELP_URI: ClassVar[str] = (
        "https://msticpy.readthedocs.io/en/latest/DataEnrichment.html"
    )

    PROVIDERS: ClassVar[dict[str, tuple[str, str]]] = {}
    CUSTOM_PROVIDERS: ClassVar[dict[str, type[Provider]]]

    PACKAGE: ClassVar[str] = ""

    def __init__(
        self: Lookup,
        providers: list[str] | None = None,
        *,
        primary_providers: list[Provider] | None = None,
        secondary_providers: list[Provider] | None = None,
    ) -> None:
        """
        Initialize TILookup instance.

        Parameters
        ----------
        primary_providers : Optional[list[Provider]], optional
            Primary Providers, by default None
        secondary_providers : Optional[list[Provider]], optional
            Secondary Providers, by default None
        providers: Optional[list[str]], optional
            List of provider names to load, by default all available
            providers are loaded. To see the list of available providers
            call `TILookup.list_available_providers()`.
            Note: if primary_provides or secondary_providers is specified
            This will override the providers list.

        """
        self._providers: dict[str, Provider] = {}
        self._secondary_providers: dict[str, Provider] = {}
        self._providers_to_load: list[str] | None = providers

        if primary_providers:
            for prov in primary_providers:
                self.add_provider(prov, primary=True)
        if secondary_providers:
            warnings.warn(
                "'secondary_providers' is a deprecated parameter",
                DeprecationWarning,
                stacklevel=1,
            )
            for prov in secondary_providers:
                self.add_provider(prov, primary=False)
        if not (primary_providers or secondary_providers):
            self._load_providers()

        self._all_providers = ChainMap(self._secondary_providers, self._providers)
        if is_ipython():
            nest_asyncio.apply()

    @property
    def loaded_providers(self: Self) -> dict[str, Provider]:
        """
        Return dictionary of loaded providers.

        Returns
        -------
        dict[str, TIProvider]
            [description]

        """
        return dict(self._all_providers)

    @property
    def provider_status(self: Self) -> Iterable[str]:
        """
        Return loaded provider status.

        Returns
        -------
        Iterable[str]
            List of providers and descriptions.

        """
        prim: list[str] = [
            f"{prov_name} - {prov.description} (primary)"
            for prov_name, prov in self._providers.items()
        ]
        sec: list[str] = [
            f"{prov_name} - {prov.description} (secondary)"
            for prov_name, prov in self._secondary_providers.items()
        ]
        return prim + sec

    @property
    def configured_providers(self: Self) -> list[str]:
        """
        Return a list of available providers that have configuration details present.

        Returns
        -------
        list[str]
            List of TI Provider classes.

        """
        prim_conf = list(self._providers.keys())
        sec_conf = list(self._secondary_providers.keys())

        return prim_conf + sec_conf

    def enable_provider(self: Self, providers: str | Iterable[str]) -> None:
        """
        Set the provider(s) as primary (used by default).

        Parameters
        ----------
        providers : Union[str, Iterable[str]
            Provider name or list of names.
            Use `list_available_providers()` to see the list of loaded providers.

        Raises
        ------
        ValueError
            If the provider name is not recognized.

        """
        provs_to_enable: list[str] | Iterable[str] = (
            [providers] if isinstance(providers, str) else providers
        )
        for provider in provs_to_enable:
            if provider in self._secondary_providers:
                self._providers[provider] = self._secondary_providers[provider]
                del self._secondary_providers[provider]
            elif provider not in self._providers:
                available_providers: list[str] | None = self.list_available_providers(
                    as_list=True,
                )
                if not available_providers:
                    err_msg: str = (
                        f"Unknown provider '{provider}'. No available providers."
                    )
                else:
                    err_msg = (
                        f"Unknown provider '{provider}'. Available providers:"
                        ", ".join(available_providers)
                    )
                raise ValueError(err_msg)

    def disable_provider(self: Self, providers: str | Iterable[str]) -> None:
        """
        Set the provider as secondary (not used by default).

        Parameters
        ----------
        providers : Union[str, Iterable[str]
            Provider name or list of names.
            Use `list_available_providers()` to see the list of loaded providers.

        Raises
        ------
        ValueError
            If the provider name is not recognized.

        """
        provs_to_disable: list[str] | Iterable[str] = (
            [providers] if isinstance(providers, str) else providers
        )
        for provider in provs_to_disable:
            if provider in self._providers:
                self._secondary_providers[provider] = self._providers[provider]
                del self._providers[provider]
            elif provider not in self._secondary_providers:
                available_providers: list[str] | None = self.list_available_providers(
                    as_list=True,
                )
                if not available_providers:
                    err_msg: str = (
                        f"Unknown provider '{provider}'. No available providers."
                    )
                else:
                    err_msg = (
                        f"Unknown provider '{provider}'. Available providers:"
                        ", ".join(available_providers)
                    )
                raise ValueError(err_msg)

    def set_provider_state(self: Self, prov_dict: dict[str, bool]) -> None:
        """
        Set a dict of providers to primary/secondary.

        Parameters
        ----------
        prov_dict : dict[str, bool]
            Dictionary of provider name and bool - True if enabled/primary,
            False if disabled/secondary.

        """
        for prov, state in prov_dict.items():
            if state:
                self.enable_provider(prov)
            else:
                self.disable_provider(prov)

    @classmethod
    def browse_results(
        cls: type[Self],
        data: pd.DataFrame,
        severities: list[str] | None = None,
        *,
        height: str = "300px",
    ) -> SelectItem | None:
        """
        Return TI Results list browser.

        Parameters
        ----------
        data : pd.DataFrame
            TI Results data from TIProviders
        severities : Optional[list[str]], optional
            A list of the severity classes to show.
            By default these are ['warning', 'high'].
            Pass ['information', 'warning', 'high'] to see all
            results.
        height: str, Optional
            Height of the widget

        Returns
        -------
        SelectItem
            SelectItem browser for TI Data.

        """
        if not isinstance(data, pd.DataFrame):
            logger.info("Input data is in an unexpected format.")
            return None
        return browse_results(data=data, severities=severities, height=height)

    browse: Callable[..., SelectItem | None] = browse_results

    def provider_usage(self: Self) -> None:
        """Print usage of loaded providers."""
        print("Primary providers")
        print("-----------------")
        if self._providers:
            for prov_name, prov in self._providers.items():
                print("\nProvider class: %s", prov_name)
                prov.usage()
        else:
            print("none")
        print("\nSecondary providers")
        print("-------------------")
        if self._secondary_providers:
            for prov_name, prov in self._secondary_providers.items():
                print("\nProvider class: %s", prov_name)
                prov.usage()
        else:
            print("none")

    @classmethod
    def reload_provider_settings(cls: type[Self]) -> None:
        """Reload provider settings from config."""
        reload_settings()
        logger.info(
            "Settings reloaded. Use reload_providers to update settings for loaded providers.",
        )

    def reload_providers(self: Self) -> None:
        """Reload settings and provider classes."""
        reload_settings()
        self._load_providers()

    def add_provider(
        self: Self,
        provider: Provider,
        name: str | None = None,
        *,
        primary: bool = True,
    ) -> None:
        """
        Add a provider to the current collection.

        Parameters
        ----------
        provider : TIProvider
            Provider instance
        name : str, optional
            The name to use for the provider (overrides the class name
            of `provider`)
        primary : bool, optional
            "primary" or "secondary" if False, by default "primary"

        """
        if not name:
            name = str(provider.__class__.__name__)
        if primary:
            self._providers[name] = provider
        else:
            self._secondary_providers[name] = provider

    def lookup_item(  # pylint: disable=too-many-locals, too-many-arguments #noqa: PLR0913
        self: Self,
        item: str,
        item_type: str | None = None,
        query_type: str | None = None,
        providers: list[str] | None = None,
        default_providers: list[str] | None = None,
        prov_scope: str = "primary",
        *,
        progress: bool = True,
        show_not_supported: bool = False,
        start: datetime.datetime | None = None,
        end: datetime.datetime | None = None,
    ) -> pd.DataFrame:
        """
        Lookup single item in active providers.

        Parameters
        ----------
        item : str
            item to lookup
        item_type : str, optional
            One of ItemExtract.ItemType, by default None
            If none, the Item type will be inferred
        query_type: str, optional
            The query type (e.g. rep, info, malware)
        providers: list[str]
            Explicit list of providers to use
        default_providers: Optional[list[str]], optional
            Used by pivot functions as a fallback to `providers`. If
            `providers` is specified, it will override this parameter.
        prov_scope : str, optional
            Use "primary", "secondary" or "all" providers, by default "primary"
        show_not_supported: bool
            If True, display unsupported items. Defaults to False
        start: dt.datetime
            If supported by the provider, start time for the item's validity
        end: dt.datetime
            If supported by the provider, end time for the item's validity
        progress : bool
            Use progress bar to track completion, by default True

        Returns
        -------
        pd.DataFrame
            DataFrame of results

        """
        return self.lookup_items(
            data={item: item_type},
            query_type=query_type,
            providers=providers,
            default_providers=default_providers,
            prov_scope=prov_scope,
            show_not_supported=show_not_supported,
            start=start,
            end=end,
            progress=progress,
        )

    def lookup_items(  # pylint: disable=too-many-arguments #noqa: PLR0913
        self: Self,
        data: pd.DataFrame | Mapping[str, str] | Sized,
        item_col: str | None = None,
        item_type_col: str | None = None,
        query_type: str | None = None,
        providers: list[str] | None = None,
        default_providers: list[str] | None = None,
        prov_scope: str = "primary",
        *,
        progress=True,
        show_not_supported: bool = False,
        start: datetime.datetime | None = None,
        end: datetime.datetime | None = None,
    ) -> pd.DataFrame:
        """
        Lookup a collection of items.

        Parameters
        ----------
        data : Union[pd.DataFrame, Mapping[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `item_col` parameter)
            2. Mapping (e.g. a dict) of [item, ItemType]
            3. Iterable of items - ItemTypes will be inferred
        item_col : str, optional
            DataFrame column to use for items, by default None
            ("col" and "column" are also aliases for this parameter)
        item_type_col : str, optional
            DataFrame column to use for ItemTypes, by default None
        query_type: str, optional
            The item query type (e.g. rep, info, malware)
        providers: list[str]
            Explicit list of providers to use
        default_providers: Optional[list[str]], optional
            Used by pivot functions as a fallback to `providers`. If
            `providers` is specified, it will override this parameter.
        prov_scope : str, optional
            Use "primary", "secondary" or "all" providers, by default "primary"
        show_not_supported: bool
            If True, display unsupported items. Defaults to False
        start: dt.datetime
            If supported by the provider, start time for the item's validity
        end: dt.datetime
            If supported by the provider, end time for the item's validity
        progress : bool
            Use progress bar to track completion, by default True

        Other Parameters
        ----------------
        progress : bool
            Use progress bar to track completion, by default True

        Returns
        -------
        pd.DataFrame
            DataFrame of results

        """
        return _make_sync(
            self._lookup_items_async(
                data=data,
                item_col=item_col,
                item_type_col=item_type_col,
                query_type=query_type,
                providers=providers,
                default_providers=default_providers,
                prov_scope=prov_scope,
                show_not_supported=show_not_supported,
                start=start,
                end=end,
                progress=progress,
            ),
        )

    @staticmethod
    def result_to_df(item_lookup: pd.DataFrame) -> pd.DataFrame:
        """
        Return DataFrame representation of Lookup response.

        Parameters
        ----------
        item_lookup : pd.DataFrame
            Output from `lookup_item`

        Returns
        -------
        pd.DataFrame
            The response as a DataFrame with a row for each
            provider response.

        """
        if not isinstance(item_lookup, pd.DataFrame):
            err_msg: str = f"DataFrame was expected, but {type(item_lookup)} received."
            raise TypeError(err_msg)
        return item_lookup

    async def _lookup_items_async(  # pylint: disable=too-many-locals, too-many-arguments #noqa: PLR0913
        self: Self,
        data: pd.DataFrame | Mapping[str, str] | Sized,
        item_col: str | None = None,
        item_type_col: str | None = None,
        query_type: str | None = None,
        providers: list[str] | None = None,
        default_providers: list[str] | None = None,
        prov_scope: str = "primary",
        *,
        show_not_supported: bool = False,
        show_bad_item: bool = False,
        progress: bool = True,
        col: str | None = None,
        column: str | None = None,
        start: datetime.datetime | None = None,
        end: datetime.datetime | None = None,
    ) -> pd.DataFrame:
        """Lookup items async."""
        del start, end
        item_col = item_col or col or column
        selected_providers: dict[str, Any] = self._select_providers(
            providers or default_providers,
            prov_scope,
        )
        if not selected_providers:
            raise MsticpyUserConfigError(
                self._NO_PROVIDERS_MSG,
                title="No Threat Intel Provider configuration found.",
                help_uri=self._HELP_URI,
            )

        event_loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        result_futures: list[Any] = []
        provider_names: list[str] = []

        prog_counter = ProgressCounter(
            total=len(data) * len(selected_providers),
        )

        # create a list of futures/tasks to await
        for prov_name, provider in selected_providers.items():
            provider_names.append(prov_name)
            result_futures.append(
                provider.lookup_items_async(
                    data=data,
                    item_col=item_col,
                    item_type_col=item_type_col,
                    query_type=query_type,
                    prog_counter=prog_counter if progress else None,
                ),
            )
        if progress:
            # Create a task for tqdm
            prog_task: asyncio.Task = event_loop.create_task(
                self._track_completion(prog_counter),
            )
        # collect the return values of the tasks
        results: list[pd.DataFrame] = await asyncio.gather(*result_futures)
        # cancel the progress task if results have completed.
        if progress:
            prog_task.cancel()
        return self._combine_results(
            results,
            provider_names,
            show_not_supported=show_not_supported,
            show_bad_item=show_bad_item,
        )

    def lookup_items_sync(  # pylint: disable=too-many-arguments, too-many-locals #noqa: PLR0913
        self: Self,
        data: pd.DataFrame | Mapping[str, str] | Iterable[str],
        item_col: str | None = None,
        item_type_col: str | None = None,
        query_type: str | None = None,
        providers: list[str] | None = None,
        default_providers: list[str] | None = None,
        prov_scope: str = "primary",
        *,
        col: str | None = None,
        column: str | None = None,
        show_not_supported: bool = False,
        show_bad_item: bool = False,
    ) -> pd.DataFrame:
        """
        Lookup a collection of items.

        Parameters
        ----------
        data : Union[pd.DataFrame, Mapping[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `item_col` parameter)
            2. Mapping (e.g. a dict) of [item, ItemType]
            3. Iterable of items - ItemTypes will be inferred
        item_col : str, optional
            DataFrame column to use for items, by default None
            ("col" and "column" are also aliases for this parameter)
        item_type_col : str, optional
            DataFrame column to use for ItemTypes, by default None
        query_type: str, optional
            The item query type (e.g. rep, info, malware)
        providers: list[str]
            Explicit list of providers to use
        default_providers: Optional[list[str]], optional
            Used by pivot functions as a fallback to `providers`. If
            `providers` is specified, it will override this parameter.
        prov_scope : str, optional
            Use "primary", "secondary" or "all" providers, by default "primary"
        col: str, Optional
            Name of the column holding the data
        column: str, Optional
            Name of the column holding the data
        show_not_supported: bool, Optional
            Set to True to include unsupported items in the result DF.
            Defaults to False
        show_bad_item: bool, Optional
            Set to True to include invalid items in the result DF.
            Defaults to False

        Returns
        -------
        pd.DataFrame
            DataFrame of results

        """
        item_col = item_col or col or column

        selected_providers: dict[str, Provider] = self._select_providers(
            providers or default_providers,
            prov_scope,
        )
        if not selected_providers:
            raise MsticpyUserConfigError(
                self._NO_PROVIDERS_MSG,
                title="No Threat Intel Provider configuration found.",
                help_uri=self._HELP_URI,
            )

        results: list[pd.DataFrame] = []
        provider_names: list[str] = []
        for prov_name, provider in selected_providers.items():
            provider_names.append(prov_name)
            results.append(
                provider.lookup_items(
                    data=data,
                    item_col=item_col,
                    item_type_col=item_type_col,
                    query_type=query_type,
                ),
            )
        return self._combine_results(
            results,
            provider_names,
            show_not_supported=show_not_supported,
            show_bad_item=show_bad_item,
        )

    @staticmethod
    async def _track_completion(prog_counter: ProgressCounter) -> None:
        total: float = await prog_counter.get_remaining()
        with tqdm(total=total, unit="obs", desc="Observables processed") as prog_bar:
            try:
                last_remaining: float = total
                while last_remaining:
                    new_remaining: float = await prog_counter.get_remaining()
                    incr: float = last_remaining - new_remaining
                    if incr:
                        prog_bar.update(incr)
                    last_remaining = new_remaining
                    await asyncio.sleep(0)
            except asyncio.CancelledError:
                # make progress bar get to 100% on cancel
                final_remaining: float = await prog_counter.get_remaining()
                if final_remaining:
                    prog_bar.update(total - final_remaining)

    @property
    def available_providers(self: Self) -> list[str]:
        """
        Return a list of builtin and plugin providers.

        Returns
        -------
        list[str]
            List of TI Provider classes.

        """
        return list(self.PROVIDERS) + list(self.CUSTOM_PROVIDERS)

    @classmethod
    def list_available_providers(
        cls: type[Self],
        *,
        show_query_types: bool = False,
        as_list: bool = False,
    ) -> list[str] | None:
        """
        Print a list of builtin providers with optional usage.

        Parameters
        ----------
        show_query_types : bool, optional
            Show query types supported by providers, by default False
        as_list : bool, optional
            Return list of providers instead of printing to stdout.
            Note: if you specify `show_query_types` this will be printed
            irrespective of this parameter setting.

        Returns
        -------
        Optional[list[str]]
            A list of provider names (if `return_list=True`)

        """
        providers: list[str] = []
        for provider_name in cls.PROVIDERS:
            provider_class: type[Provider] = cls.import_provider(provider_name)
            if not as_list:
                logger.info(provider_name)
            providers.append(provider_name)
            if show_query_types and provider_class:
                provider_class.usage()

        return providers if as_list else None

    @classmethod
    def import_provider(cls: type[Self], provider: str) -> type[Provider]:
        """Import provider class."""
        mod_name, cls_name = cls.PROVIDERS.get(provider, (None, None))

        if not (mod_name and cls_name):
            if hasattr(cls, "CUSTOM_PROVIDERS") and provider in cls.CUSTOM_PROVIDERS:
                return cls.CUSTOM_PROVIDERS[provider]
            err_msg: str = (
                f"No provider named '{provider}'. Possible values are: "
                ", ".join(list(cls.PROVIDERS) + list(cls.CUSTOM_PROVIDERS))
            )
            raise LookupError(err_msg)

        imp_module: ModuleType = importlib.import_module(
            f"msticpy.context.{cls.PACKAGE}.{mod_name}",
            package="msticpy",
        )
        return getattr(imp_module, cls_name)

    def _load_providers(
        self: Self,
        *,
        providers: str = "Providers",
    ) -> None:
        """Load provider classes based on config."""
        prov_settings: dict[str, ProviderSettings] = get_provider_settings(providers)
        for provider_name, settings in prov_settings.items():
            if (
                self._providers_to_load is not None
                and provider_name not in self._providers_to_load
            ) or settings.provider == "--no-load--":
                continue
            try:
                prov_name: str = settings.provider or provider_name
                provider_class: type[Provider] = self.import_provider(prov_name)
            except LookupError:
                warnings.warn(
                    f"Could not find provider class for {provider_name} "
                    f"in config section '{provider_name}'. "
                    f"Provider class name in config is '{settings.provider}'",
                    stacklevel=2,
                )
                prov_name = provider_name
                continue

            # instantiate class sending args from settings to init
            try:
                # filter out any args that are not valid for the provider
                provider_args = {
                    key: value
                    for key, value in settings.args.items()
                    if key in _HTTP_PROVIDER_LEGAL_KWARGS
                }
                provider_instance: Provider = provider_class(**(provider_args))
            except MsticpyConfigError as mp_ex:
                # If the TI Provider didn't load, raise an exception
                err_msg: str = (
                    f"Could not load Provider {provider_name} {mp_ex.args}"
                    "To avoid loading this provider please use the 'providers' parameter"
                    " to specify which providers to load."
                )
                raise MsticpyUserConfigError(
                    err_msg,
                    title="Provider configuration error",
                    help_uri=self._HELP_URI,
                ) from mp_ex

            # set the description from settings, if one is provided, otherwise
            # use class docstring.
            provider_instance.description = (
                settings.description or provider_instance.__doc__
            )

            self.add_provider(
                provider=provider_instance,
                name=prov_name,
                primary=settings.primary,
            )

    def _select_providers(
        self: Self,
        providers: list[str] | None = None,
        prov_scope: str = "primary",
    ) -> dict[str, Provider]:
        """
        Return required subset of providers.

        Parameters
        ----------
        providers : list[str], optional
            Explicit list of provider names, by default None
        prov_scope : str, optional
            Provider scope, by default "primary"
            Other values are "all" and "secondary"

        Returns
        -------
        dict[str, TIProvider]
            Dictionary of provider names and instances.

        """
        if providers:
            return {
                prov_name: prov
                for prov_name, prov in self._all_providers.items()
                if prov_name in providers
            }
        if prov_scope == "all":
            return dict(self._all_providers)
        if prov_scope == "primary":
            return self._providers
        return self._secondary_providers

    @staticmethod
    def _combine_results(
        results: Iterable[pd.DataFrame],
        provider_names: list[str],
        *,
        show_not_supported: bool = False,
        show_bad_item: bool = False,
    ) -> pd.DataFrame:
        """Combine dataframe results into single DF."""
        result_list: list[pd.DataFrame] = []
        for prov_name, provider_result in zip(provider_names, results):
            if provider_result is None or provider_result.empty:
                continue
            result: pd.DataFrame = provider_result.copy()
            if not show_not_supported:
                result = result[result["Status"] != LookupStatus.NOT_SUPPORTED.value]
            if not show_bad_item:
                result = result[result["Status"] != LookupStatus.BAD_FORMAT.value]
            result["Provider"] = prov_name
            result_list.append(result)

        if not result_list:
            logger.info("No Item matches")
        return pd.concat(result_list, sort=False) if result_list else pd.DataFrame()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""MSTICPy core pandas accessor methods."""
from typing import Any, Dict, List, Mapping, Union

import pandas as pd

from .._version import VERSION
from ..context.ip_utils import get_whois_df
from ..data.data_obfus import mask_df
from ..transform.base64unpack import unpack_df
from ..transform.iocextract import IoCExtract
from ..transform.network import df_to_networkx
from ..transform.proc_tree_builder import ProcSchema, build_process_tree

__version__ = VERSION
__author__ = "Ian Hellen"


@pd.api.extensions.register_dataframe_accessor("mp")
class MsticpyCoreAccessor:
    """Msticpy pandas accessor for core functions."""

    def __init__(self, pandas_obj):
        """Initialize the extension."""
        self._df = pandas_obj
        self._ioc = IoCExtract()

        self._data_viewer_class: Any = None

    def b64extract(self, column: str, **kwargs) -> pd.DataFrame:
        """
        Base64-decode strings taken from a pandas dataframe.

        Parameters
        ----------
        data : pd.DataFrame
            dataframe containing column to decode
        column : str
            Name of dataframe text column
        trace : bool, optional
            Show additional status (the default is None)
        utf16 : bool, optional
            Attempt to decode UTF16 byte strings

        Returns
        -------
        pd.DataFrame
            Decoded string and additional metadata in dataframe

        Notes
        -----
        Items that decode to utf-8 or utf-16 strings will be returned as decoded
        strings replaced in the original string. If the encoded string is a
        known binary type it will identify the file type and return the hashes
        of the file. If any binary types are known archives (zip, tar, gzip) it
        will unpack the contents of the archive.
        For any binary it will return the decoded file as a byte array, and as a
        printable list of byte values.

        The columns of the output DataFrame are:

        - decoded string: this is the input string with any decoded sections
          replaced by the results of the decoding
        - reference : this is an index that matches an index number in the
          decoded string (e.g. <<encoded binary type=pdf index=1.2').
        - original_string : the string prior to decoding - file_type : the type
          of file if this could be determined
        - file_hashes : a dictionary of hashes (the md5, sha1 and sha256 hashes
          are broken out into separate columns)
        - input_bytes : the binary image as a byte array
        - decoded_string : printable form of the decoded string (either string
          or list of hex byte values)
        - encoding_type : utf-8, utf-16 or binary
        - md5, sha1, sha256 : the respective hashes of the binary file_type,
          file_hashes, input_bytes, md5, sha1, sha256 will be null if this item is
          decoded to a string
        - src_index - the index of the source row in the input
          frame.

        """
        return unpack_df(data=self._df, column=column, **kwargs)

    def ioc_extract(self, columns: List[str], **kwargs) -> pd.DataFrame:
        """
        Extract IoCs from either a pandas DataFrame.

        Parameters
        ----------
        columns : list
            The list of columns to use as source strings,

        Other Parameters
        ----------------
        ioc_types : list, optional
            Restrict matching to just specified types.
            (default is all types)
        include_paths : bool, optional
            Whether to include path matches (which can be noisy)
            (the default is false - excludes 'windows_path'
            and 'linux_path'). If `ioc_types` is specified
            this parameter is ignored.

        Returns
        -------
        pd.DataFrame
            DataFrame of observables

        Notes
        -----
        Extract takes a pandas DataFrame as input.
        The results will be returned as a new
        DataFrame with the following columns:
        - IoCType: the mnemonic used to distinguish different IoC Types
        - Observable: the actual value of the observable
        - SourceIndex: the index of the row in the input DataFrame from
        which the source for the IoC observable was extracted.

        IoCType Pattern selection
        The default list is:  ['ipv4', 'ipv6', 'dns', 'url',
        'md5_hash', 'sha1_hash', 'sha256_hash'] plus any
        user-defined types.
        'windows_path', 'linux_path' are excluded unless `include_paths`
        is True or explicitly included in `ioc_paths`.

        """
        return self._ioc.extract_df(data=self._df, columns=columns, **kwargs)

    def build_process_tree(
        self,
        schema: Union[ProcSchema, Dict[str, Any]] = None,
        show_summary: bool = False,
        debug: bool = False,
    ) -> pd.DataFrame:
        """
        Build process trees from the process events.

        Parameters
        ----------
        schema : Union[ProcSchema, Dict[str, Any]], optional
            The column schema to use, by default None.
            If supplied as a dict it must include definitions for the
            required fields in the ProcSchema class
            If None, then the schema is inferred
        show_summary : bool
            Shows summary of the built tree, default is False.
        debug : bool
            If True produces extra debugging output,
            by default False

        Returns
        -------
        pd.DataFrame
            Process tree dataframe.

        See Also
        --------
        ProcSchema

        """
        return build_process_tree(
            procs=self._df, schema=schema, show_summary=show_summary, debug=debug
        )

    def to_graph(self, **kwargs):
        """
        Create a networkx graph from a DataFrame.

        Parameters
        ----------
        source_col : str
            Column for source nodes.
        target_col : str
            Column for target nodes.
        source_attrs : Optional[List[str]], optional
            Optional list of columns to use as source node attributes, by default None
        target_attrs : Optional[List[str]], optional
            Optional list of columns to use as target node attributes, by default None
        edge_attrs : Optional[List[str]], optional
            Optional list of columns to use as edge node attributes, by default None
        graph_type : str
            "graph" or "digraph" (for nx.DiGraph)

        Returns
        -------
        nx.Graph
            The networkx graph object

        """
        return df_to_networkx(self._df, **kwargs)

    def mask(
        self, column_map: Mapping[str, Any] = None, use_default: bool = True
    ) -> pd.DataFrame:
        """
        Obfuscate the data in columns of a pandas dataframe.

        Parameters
        ----------
        data : pd.DataFrame
            dataframe containing column to obfuscate
        column_map : Mapping[str, Any], optional
            Custom column mapping, by default None
        use_default: bool
            If True use the built-in map (adding any custom
            mappings to this dictionary)

        Returns
        -------
        pd.DataFrame
            Obfuscated dataframe

        """
        return mask_df(data=self._df, column_map=column_map, use_default=use_default)

    def whois(self, ip_column, **kwargs):
        """
        Extract IoCs from either a pandas DataFrame.

        Parameters
        ----------
        ip_column : str
            Column name of IP Address to look up.

        Other Parameters
        ----------------
        asn_col : str, optional
            Name of the output column for ASN description,
            by default "ASNDescription"
        ip_column : str, optional
            Name of the output column for full whois data,
            by default "WhoIsData"
        show_progress : bool, optional
            Show progress for each query, by default False

        Returns
        -------
        pd.DataFrame
            Output DataFrame with results in added columns.

        """
        return get_whois_df(data=self._df, ip_column=ip_column, **kwargs)

    def view(self, **kwargs):
        """
        Initialize the DataViewer class.

        Parameters
        ----------
        data : pd.DataFrame
            The DataFrame to view
        selected_cols : List[str], optional
            Initial subset of columns to show, by default None (all cols)

        Other Parameters
        ----------------
        selectable : Union[bool, str], optional
            Whether rows should be selectable, by default "checkbox"
        show_index : bool, optional
            If True show the DataFrame index as a column, by default  True.
        show_tenant_id : bool, optional
            If True show the TenantId column, by default  True.
        max_col_width : int, optional
            Sets the maximum column width to display, by default 500
        detail_cols : List[str]
            List of columns for which details are displayed in collapsible
            field beneath each table row.
        kwargs :
            Other keyword arguments are passed to the panel
            Tabulator control.

        Notes
        -----
        Main attributes:
        value - original dataframe
        selected - indexes of currently selected rows
        selected_dataframe - currently selected rows
        current_view - current dataframe after filtering and sorting
        selection - indexes of currently selected rows

        See Also
        --------
        Tabulator - https://panel.holoviz.org/reference/widgets/Tabulator.html

        """
        if self._data_viewer_class is None:
            try:
                # pylint: disable=import-outside-toplevel
                from ..vis.data_viewer_panel import DataViewer
            except ImportError:
                print("This component needs the panel package.")
                return self._df
            self._data_viewer_class = DataViewer
        return self._data_viewer_class(data=self._df, **kwargs)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Module for TILookup classes.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from __future__ import annotations

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, ClassVar, Iterable

from typing_extensions import Self

from ..._version import VERSION
from ...common.utility import export
from ..provider_base import PivotProvider, Provider
from .result_severity import ResultSeverity

if TYPE_CHECKING:
    import pandas as pd

    from ...init.pivot import Pivot
    from ...init.pivot_core.pivot_register import PivotRegistration

logger: logging.Logger = logging.getLogger(__name__)

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class TIProvider(Provider):
    """Abstract base class for Threat Intel providers."""

    _QUERIES: ClassVar[dict[str, Any]] = {}

    def _check_item_type(
        self: Self,
        item: str,
        item_type: str | None = None,
        query_subtype: str | None = None,
    ) -> dict:
        """
        Check Item Type and cleans up item.

        Parameters
        ----------
        item : str
            item
        item_type : str, optional
            item type, by default None
        query_subtype : str, optional
            Query sub-type, if any, by default None

        Returns
        -------
        Dict
            Dict result with resolved type and pre-processed
            item.
            Status is none-zero on failure.

        """
        return self._check_ioc_type(
            item,
            item_type,
            query_subtype,
        )

    def _check_ioc_type(
        self: Self,
        ioc: str,
        ioc_type: str | None = None,
        query_subtype: str | None = None,
    ) -> dict:
        """
        Check Ioc Type and cleans up ioc.

        Parameters
        ----------
        ioc : str
            IoC
        ioc_type : str, optional
            IoC type, by default None
        query_subtype : str, optional
            Query sub-type, if any, by default None

        Returns
        -------
        Dict
            Dict result with resolved type and pre-processed
            Ioc.
            Status is none-zero on failure.

        """
        result: dict[str, Any] = super()._check_item_type(
            item=ioc,
            item_type=ioc_type,
            query_subtype=query_subtype,
        )
        result["Ioc"] = result.pop("Item")
        result["IocType"] = result.pop("ItemType")
        result["SafeIoc"] = result.pop("SanitizedValue")
        result["Severity"] = ResultSeverity.information.name
        return result

    def lookup_item(
        self: Self,
        item: str,
        item_type: str | None = None,
        query_type: str | None = None,
    ) -> pd.DataFrame:
        """
        Lookup a single item.

        Parameters
        ----------
        item : str
            Item value to lookup
        item_type : str, optional
            The Type of the value to lookup, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the item_value
            will be returned.

        Returns
        -------
        pd.DataFrame
            The lookup result:
            result - Positive/Negative,
            details - Lookup Details (or status if failure),
            raw_result - Raw Response
            reference - URL of the item

        Raises
        ------
        NotImplementedError
            If attempting to use an HTTP method or authentication
            protocol that is not supported.

        Notes
        -----
        Note: this method uses memoization (lru_cache) to cache results
        for a particular observable to try avoid repeated network calls for
        the same item.

        """
        return self.lookup_ioc(
            ioc=item,
            ioc_type=item_type,
            query_type=query_type,
        )

    @abstractmethod
    def parse_results(self: Self, response: dict) -> tuple[bool, ResultSeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : Dict
            The returned data response

        Returns
        -------
        tuple[bool, ResultSeverity, Any]
            bool = positive or negative hit
            ResultSeverity = enumeration of severity
            Object with match details

        """

    @abstractmethod
    def lookup_ioc(
        self: Self,
        ioc: str,
        ioc_type: str | None = None,
        query_type: str | None = None,
    ) -> pd.DataFrame:
        """
        Lookup a single IoC observable.

        Parameters
        ----------
        ioc : str
            IoC Observable value
        ioc_type : str, optional
            IoC Type, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.

        Returns
        -------
        pd.DataFrame
            DataFrame of results.

        """

    def lookup_iocs(
        self: Self,
        data: pd.DataFrame | dict[str, str] | Iterable[str],
        ioc_col: str | None = None,
        ioc_type_col: str | None = None,
        query_type: str | None = None,
    ) -> pd.DataFrame:
        """
        Lookup collection of IoC observables.

        Parameters
        ----------
        data : Union[pd.DataFrame, dict[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `ioc_col` parameter)
            2. Dict of observable, IoCType
            3. Iterable of observables - IoCTypes will be inferred
        ioc_col : str, optional
            DataFrame column to use for observables, by default None
        ioc_type_col : str, optional
            DataFrame column to use for IoCTypes, by default None
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.

        Returns
        -------
        pd.DataFrame
            DataFrame of results.

        """
        return self.lookup_items(
            data,
            item_col=ioc_col,
            item_type_col=ioc_type_col,
            query_type=query_type,
        )

    async def lookup_iocs_async(
        self: Self,
        data: pd.DataFrame | dict[str, str] | Iterable[str],
        ioc_col: str | None = None,
        ioc_type_col: str | None = None,
        query_type: str | None = None,
    ) -> pd.DataFrame:
        """Call base async wrapper."""
        return await self._lookup_items_async_wrapper(
            data,
            item_col=ioc_col,
            item_type_col=ioc_type_col,
            query_type=query_type,
        )

    async def _lookup_iocs_async_wrapper(
        self: Self,
        data: pd.DataFrame | dict[str, str] | Iterable[str],
        ioc_col: str | None = None,
        ioc_type_col: str | None = None,
        query_type: str | None = None,
    ) -> pd.DataFrame:
        """
        Async wrapper for providers that do not implement lookup_iocs_async.

        Parameters
        ----------
        data : Union[pd.DataFrame, dict[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `obs_col` parameter)
            2. Dict of observable, IoCType
            3. Iterable of observables - IoCTypes will be inferred
        ioc_col : str, optional
            DataFrame column to use for observables, by default None
        ioc_type_col : str, optional
            DataFrame column to use for IoCTypes, by default None
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.

        Returns
        -------
        pd.DataFrame
            DataFrame of results.

        """
        return await self._lookup_items_async_wrapper(
            data,
            item_col=ioc_col,
            item_type_col=ioc_type_col,
            query_type=query_type,
        )

    @property
    def ioc_query_defs(self: Self) -> dict[str, Any]:
        """
        Return current dictionary of IoC query/request definitions.

        Returns
        -------
        dict[str, Any]
            IoC query/request definitions keyed by IoCType

        """
        return self._QUERIES

    @classmethod
    def usage(cls: type[Self]) -> None:
        """Print usage of provider."""
        print(f"{cls.__doc__} Supported query types:")
        for ioc_key in sorted(cls._QUERIES):
            ioc_key_elems: list[str] = ioc_key.split("-", maxsplit=1)
            try:
                print(f"\tioc_type={ioc_key_elems[0]}, query_type={ioc_key_elems[1]}")
            except IndexError:
                print(f"\tioc_type={ioc_key_elems[0]}")

    @staticmethod
    def resolve_ioc_type(observable: str) -> str:
        """
        Return IoCType determined by IoCExtract.

        Parameters
        ----------
        observable : str
            IoC observable string

        Returns
        -------
        str
            IoC Type (or unknown if type could not be determined)

        """
        return TIProvider.resolve_item_type(observable)


class TIPivotProvider(PivotProvider):
    """A class which provides TI pivot functions and a means of registering them."""

    @abstractmethod
    def register_pivots(
        self: Self,
        pivot_reg: type[PivotRegistration],
        pivot: Pivot,
    ) -> None:
        """
        Register pivot functions for the TI Provider.

        Parameters
        ----------
        pivot_reg : PivotRegistration
            Pivot registration settings.
        pivot : Pivot
            Pivot library instance

        """



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for common timeline functions."""
from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union

import pandas as pd
from bokeh.models import (  # type: ignore[attr-defined]
    ColumnDataSource,
    DatetimeTickFormatter,
    GestureTool,
    Label,
    LayoutDOM,
    Range,
    RangeTool,
    Title,
)

# pylint: disable=no-name-in-module
from bokeh.palettes import Palette, viridis

# pylint: enable=no-name-in-module
from bokeh.plotting import figure

try:
    from bokeh.plotting import Figure  # type: ignore
except ImportError:
    Figure = LayoutDOM
from pandas.api.types import is_datetime64_any_dtype
from pandas.errors import OutOfBoundsDatetime

from .._version import VERSION
from ..common.exceptions import MsticpyParameterError
from ..common.utility import export
from .figure_dimension import bokeh_figure

# pylint: enable=unused-import

__version__ = VERSION
__author__ = "Ian Hellen"

TIMELINE_HELP = (
    "https://msticpy.readthedocs.io/en/latest/msticpy.vis.html"
    "#msticpy.vis.timeline.{plot_type}"
)

# wrap figure function to handle v2/v3 parameter renaming
figure = bokeh_figure(figure)  # type: ignore[assignment, misc]


@export
def check_df_columns(
    data: pd.DataFrame, req_columns: List[str], help_uri: str, plot_type: str
):
    """
    Check that specified columns are in the DataFrame.

    Parameters
    ----------
    data : pd.DataFrame
        [description]
    req_columns : List[str]
        [description]
    help_uri : str
        [description]
    plot_type : str
        [description]

    Raises
    ------
    MsticpyParameterError
        If one or more columns not found in `data`

    """
    missing_cols = set(req_columns) - set(data.columns)  # type: ignore
    if missing_cols:
        raise MsticpyParameterError(
            title="Columns not found in DataFrame",
            help_uri=help_uri.format(plot_type=plot_type),
            parameter=missing_cols,
        )


def create_data_grouping(
    data: pd.DataFrame,
    source_columns: List[str],
    time_column: str,
    group_by: Optional[str],
    color: str,
) -> Tuple[pd.DataFrame, pd.DataFrame, Set[str], int]:
    """
    Group input data and add indexes and tooltips.

    Parameters
    ----------
    data : pd.DataFrame
        Input dataframe
    source_columns : List[str]
        Tooltip columns
    time_column : str
        Name of the time column
    group_by : str
        Name of the group_by column
    color : str
        Plot color

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame, Set[str], int]
        Tuple of grouped dataframe, group color dataframe,
        set of tooltip columns, the number of groups created.

    """
    data_columns = get_def_source_cols(data, source_columns)
    # If the time column not explicitly specified in source_columns, add it
    data_columns.add(time_column)
    tool_tip_columns = data_columns.copy()
    # create group frame so that we can color each group separately
    if group_by:
        group_count_df = (
            data[[group_by, time_column]]
            .groupby(group_by)
            .count()
            .reset_index()
            .rename(columns={time_column: "count"})
        )
        group_count_df["y_index"] = group_count_df.index

        # Shift the Viridis palette so we lose the top, harder-to-see colors
        series_count = len(group_count_df)
        colors, palette_size = get_color_palette(series_count)
        group_count_df["color"] = group_count_df.apply(
            lambda x: colors[x.y_index % palette_size], axis=1
        )
        # re-join with the original data
        data_columns.update([group_by, "y_index", "color"])
        clean_data = data.drop(columns=["y_index", "color"], errors="ignore")
        graph_df = clean_data.merge(group_count_df, on=group_by)[list(data_columns)]
    else:
        graph_df = data[list(data_columns)].copy()
        graph_df["color"] = color
        graph_df["y_index"] = 1
        series_count = 1
        group_count_df = None
    return graph_df, group_count_df, tool_tip_columns, series_count  # type: ignore


def get_def_source_cols(data: pd.DataFrame, source_columns: Iterable[str]) -> Set[str]:
    """Get default set of columns (backward compat)."""
    if not source_columns:
        return (
            {"NewProcessName", "EventID", "CommandLine"}
            if all(
                col in data.columns
                for col in ["NewProcessName", "EventID", "CommandLine"]
            )
            else set()
        )

    return set(source_columns)


def get_color_palette(series_count: int) -> Tuple[Palette, int]:
    """Return palette based on series size."""
    palette_size = min(256, series_count + series_count // 5)
    return viridis(palette_size), palette_size


def set_axes_and_grids(
    data: pd.DataFrame, plot: Figure, show_yaxis: bool, ygrid: bool, xgrid: bool
):
    """Set the axes visibility and grids according to parameters."""
    plot.yaxis.visible = show_yaxis
    if show_yaxis and data:
        y_labels = {ser_def["y_index"]: str(lbl) for lbl, ser_def in data.items()}
        plot.yaxis.major_label_overrides = y_labels
    if ygrid:
        plot.ygrid.minor_grid_line_color = "navy"
        plot.ygrid.minor_grid_line_alpha = 0.1
        plot.ygrid.grid_line_color = "navy"
        plot.ygrid.grid_line_alpha = 0.3
    else:
        plot.ygrid.grid_line_color = None
    if xgrid:
        plot.xgrid.minor_grid_line_color = "navy"
        plot.xgrid.minor_grid_line_alpha = 0.3
    else:
        plot.xgrid.grid_line_color = None


def get_time_bounds(
    min_time: pd.Timestamp, max_time: pd.Timestamp
) -> Tuple[pd.Timestamp, pd.Timestamp, pd.Timestamp, pd.Timestamp]:
    """Return start and end range, coping with out-of-bounds error."""
    try:
        start_range = min_time - ((max_time - min_time) * 0.1)
        end_range = max_time + ((max_time - min_time) * 0.1)
    except OutOfBoundsDatetime:
        min_time = min_time.to_pydatetime()  # type: ignore
        max_time = max_time.to_pydatetime()  # type: ignore
        start_range = min_time - ((max_time - min_time) * 0.1)
        end_range = max_time + ((max_time - min_time) * 0.1)
    return start_range, end_range, min_time, max_time


def create_tool_tips(
    data: Union[pd.DataFrame, Dict[str, pd.DataFrame]], columns: Iterable[str]
) -> Dict[str, Any]:
    """Create formatting for tool tip columns."""
    formatters: Dict[str, str] = {}
    # if this is a dict we need to unpack each dataframe and process
    # the tooltip columns for all of the data sets.
    if isinstance(data, dict):
        tool_tip_dict = {}
        for data_set in data.values():
            data_df = data_set.get("data", {})  # type: ignore
            for col in columns:
                disp_col, col_tooltip, col_fmt = _get_datetime_tooltip(col, data_df)  # type: ignore
                tool_tip_dict[disp_col] = col_tooltip
                formatters.update(col_fmt)
        return {"tooltips": list(tool_tip_dict.items()), "formatters": formatters}

    # If just a dataframe we just process the columns against this
    tool_tip_items = []
    for col in columns:
        disp_col, col_tooltip, col_fmt = _get_datetime_tooltip(col, data)
        tool_tip_items.append((disp_col, col_tooltip))
        formatters.update(col_fmt)

    return {"tooltips": tool_tip_items, "formatters": formatters}


def _get_datetime_tooltip(
    col: str, dataset: pd.DataFrame
) -> Tuple[str, str, Dict[str, str]]:
    """Return tooltip and formatter entries for column."""
    if " " in col:
        disp_col = col.replace(" ", "_")
        tt_col = f"{{{col}}}"
    else:
        disp_col = tt_col = col
    if col in dataset and is_datetime64_any_dtype(dataset[col]):
        col_tooltip = f"@{tt_col}{{%F %T.%3N}}"
        col_fmt: Dict[Any, Any] = {f"@{tt_col}": "datetime"}
    else:
        col_tooltip = f"@{tt_col}"
        col_fmt = {}
    return disp_col, col_tooltip, col_fmt


def calc_auto_plot_height(group_count: int) -> int:
    """Dynamic calculation of plot height."""
    ht_per_row = 25 if group_count > 15 else 40
    return max(ht_per_row * group_count, 300)


# pylint: disable=too-many-arguments, invalid-name, too-many-locals
def create_range_tool(
    data: pd.DataFrame,
    min_time: pd.Timestamp,
    max_time: pd.Timestamp,
    plot_range: Range,
    width: int,
    height: int,
    time_column: str = None,
    y: str = "y_index",
) -> Figure:
    """Create plot bar to act as as range selector."""
    ext_min = min_time - ((max_time - min_time) * 0.15)
    ext_max = max_time + ((max_time - min_time) * 0.15)
    plot_height = max(120, int(height * 0.20))
    rng_select = figure(
        x_range=(ext_min, ext_max),
        title="Range Selector",
        x_axis_type="datetime",
        y_axis_type=None,
        tools="",
        toolbar_location=None,
        height=plot_height,
        width=width,
    )

    help_str = (
        "Drag the middle or edges of the selection box to change "
        + "the range in the main chart"
    )
    rng_select.add_layout(
        Title(text=help_str, align="right", text_font_size="10px"), "below"
    )
    rng_select.xaxis[0].formatter = get_tick_formatter()
    if isinstance(data, dict):
        for _, series_def in data.items():
            rng_select.circle(
                x=series_def["time_column"],
                y=y,
                color=series_def["color"],
                source=series_def["source"],
            )
    elif isinstance(data, pd.DataFrame):
        rng_select.circle(
            x=time_column, y=y, color="blue", source=ColumnDataSource(data)
        )

    range_tool = RangeTool(x_range=plot_range)
    range_tool.overlay.fill_color = "navy"  # type: ignore
    range_tool.overlay.fill_alpha = 0.2  # type: ignore
    rng_select.ygrid.grid_line_color = None
    rng_select.add_tools(range_tool)
    if isinstance(range_tool, GestureTool):
        rng_select.toolbar.active_multi = range_tool  # type: ignore
    return rng_select


# pylint: enable=too-many-arguments


def plot_ref_line(
    plot: LayoutDOM,
    ref_time: datetime,
    ref_text: str = "Ref time",
    series_count: int = 1,
    index: int = 0,
):
    """Add a reference marker line and label at `ref_time`."""
    ref_label_tm = pd.Timestamp(ref_time)
    index = min(index, series_count)
    plot.line(
        x=[ref_label_tm, ref_label_tm],
        y=[0, series_count],
        line_width=1,
        line_color="red",
        line_dash="dashed",
        line_alpha=0.5,
    )
    ref_label = Label(
        x=ref_label_tm,
        y=0,
        y_offset=10 + (10 * index),
        x_units="data",
        y_units="data",
        text=f"< {ref_text}",
        text_font_size="8pt",
        text_alpha=0.5,
        border_line_color="red",
        border_line_alpha=0.3,
        background_fill_color="white",
        background_fill_alpha=0.3,
    )

    plot.add_layout(ref_label)


def plot_ref_events(
    plot: Figure,
    time_col: str,
    group_count: int,
    ref_events: Optional[pd.DataFrame] = None,
    ref_col: Optional[str] = None,
    ref_times: Optional[List[Tuple[datetime, str]]] = None,
):
    """Plot reference lines/labels."""
    if ref_events is not None:
        if isinstance(ref_events, pd.Series):
            ref_events = pd.DataFrame(ref_events)
        for idx, event in enumerate(ref_events.itertuples()):
            evt_time = event._asdict()[time_col]
            evt_label = (
                event._asdict()[ref_col] if ref_col else f"reference {event.Index}"
            )
            plot_ref_line(
                plot=plot,
                ref_time=evt_time,
                ref_text=evt_label,
                series_count=group_count,
                index=idx,
            )
    elif ref_times:
        for idx, (evt_time, evt_label) in enumerate(ref_times):
            evt_label = evt_label or f"reference {idx}"
            plot_ref_line(
                plot=plot,
                ref_time=evt_time,
                ref_text=evt_label,
                series_count=group_count,
                index=idx,
            )


def get_ref_event_time(**kwargs) -> Tuple[Optional[Any], Union[Any, str]]:
    """Extract the reference time from kwargs."""
    ref_alert = kwargs.get("alert", None)
    if ref_alert is not None:
        ref_event = ref_alert
        ref_label = "Alert time"
    else:
        ref_event = kwargs.get("ref_event", None)
        ref_label = "Event time"

    if ref_event is not None:
        if isinstance(ref_event, pd.DataFrame):
            ref_event = ref_event.iloc[0]
        ref_time = getattr(ref_event, "StartTimeUtc", None)
        if not ref_time:
            ref_time = getattr(ref_event, "TimeGenerated", None)
    else:
        ref_time = kwargs.get("ref_time", None)
        ref_label = "Ref time"
    return ref_time, kwargs.get("ref_label", ref_label)  # type: ignore


def get_tick_formatter() -> DatetimeTickFormatter:
    """Return tick formatting for different zoom levels."""
    # '%H:%M:%S.%3Nms
    tick_format = DatetimeTickFormatter()
    tick_format.days = "%m-%d %H:%M"  # type: ignore
    tick_format.hours = "%H:%M:%S"  # type: ignore
    tick_format.minutes = "%H:%M:%S"  # type: ignore
    tick_format.seconds = "%H:%M:%S"  # type: ignore
    tick_format.milliseconds = "%H:%M:%S.%3N"  # type: ignore
    return tick_format



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Pivot function hierarchy attribute class."""
from ..._version import VERSION
from ...common.data_types import ObjectContainer

__version__ = VERSION
__author__ = "Ian Hellen"


class PivotContainer(ObjectContainer):
    """Empty class used to create hierarchical attributes."""

    def __repr__(self):
        """Return list of attributes."""
        repr_list = []
        for name, obj in sorted(self.__dict__.items()):
            if isinstance(obj, ObjectContainer):
                repr_list.append(f"{name} (container)")
            elif hasattr(obj, "pivot_properties"):
                repr_list.append(f"{name} (pivot function)")
            elif not name.startswith("_"):
                repr_list.append(f"{name} ({type(obj).__name__})")
        return "\n".join(repr_list)

    def __call__(self, *args, **kwargs):
        """Return list of attributes or help."""
        if args or kwargs:
            print("This attribute is a container, not a function.")
            print("Items in this container:")
        print(repr(self))



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""SOC Entity sub-package."""

# flake8: noqa: F401
from .incident import Incident
from .sentinel_alert import SentinelAlert
